from typing import List, Optional


class ArgsQue(object):
    def __init__(self, args: List[str]):
        self.__args = args
        self.__current_index = 0
        self.__next_index = 0

    def __iter__(self):
        return self.reset()

    def __next__(self) -> str:
        if self.has_next():
            self.__current_index = self.__next_index
            self.__next_index = self.__current_index + 1
            return self.__args[self.__current_index]
        else:
            raise StopIteration

    def __len__(self) -> int:
        return len(self.__args)

    def __bool__(self) -> bool:
        return len(self.__args) > 0

    def has_next(self) -> bool:
        return self.__next_index < len(self.__args)

    def pop_current(self) -> str:
        current = self.__args.pop(self.__current_index)
        self.__next_index = self.__current_index
        return current

    def peek_current(self) -> Optional[str]:
        if self.__current_index >= len(self.__args):
            return None
        current = self.__args[self.__current_index]
        self.__next_index = self.__current_index + 1
        return current

    def pop_all(self) -> List[str]:
        all_args = self.__args
        self.__args = []
        self.reset()
        return all_args

    def reset(self) -> 'ArgsQue':
        self.__current_index = 0
        self.__next_index = 0
        return self



from typing import Union, Callable, Optional, List, Type, Any, Iterable

from .rule import SubcommandRule, PrimaryOptionRule, ParameterRule, PositionalArgumentRule, ManyArgumentsRule, \
    DefaultActionRule, FlagRule, DictionaryRule


def subcommand(
        *keywords: str,
        run: Optional[Callable[..., None]] = None,
        help: str = None,
) -> SubcommandRule:
    """
    Create Subcommand rule specification.
    Subcommand is a keyword which narrows down the context and can execute an action.
    Subcommands may have multiple levels and may build a tree.
    It's similar to 'git' syntax: 'git remote rename ...'
    Subcommand can have more subrules which are activated only when corresponding subcommand is active.
    Subrules can be added using 'has' method.
    :param keywords: keyword arguments which any of them triggers a subcommand
    :param run: optional action to be invoked when subcommand is matched
    :param help: description of the subcommand displayed in help output
    :return: new subcommand rule specification
    """
    return SubcommandRule(help, set(keywords), run)


def flag(
        *keywords: str,
        help: str = None,
        multiple: bool = False,
) -> FlagRule:
    """
    Create flag rule specification.
    Flag is a boolean parameter which is toggled by single argument
    :param keywords: keyword arguments which any of them enables flag when it occurs.
    Flag keywords may be passed using direct format: '-f' or '--flag',
    as well as by name: 'f' or 'flag', which will be evaluated to '-f' or '--flag'.
    Single character flags will get single hyphen prefix (-f),
    longer flag names will get double hyphen prefix (--flag)
    :param help: description of the flag displayed in help output
    :param multiple: whether flag is allowed to occur many times.
    Then flag has int type and stores number of its occurrences
    :return: new flag rule specification
    """
    return FlagRule(set(keywords), help, multiple)


def parameter(
        *keywords: str,
        name: str = None,
        help: str = None,
        required: bool = False,
        default: Any = None,
        type: Union[Type, Callable[[str], Any], None] = str,
        choices: Union[Iterable[Any], Callable[..., List[Any]]] = None,
        strict_choices: bool = False,
        multiple: bool = False,
) -> ParameterRule:
    """
    Create parameter rule specification.
    Parameter is a named value, which will be injected to triggered action by its name.
    Shell syntax for settings parameter value is following:
    '--parameter-name value' or '--parameter-name=value'
    The parameters may be later referenced by its explicit name or keywords
    (in lowercase format without hyphen prefix and with underscores instead of dashes,
    e.g. '--paramater-name' will be injected as 'parameter_name')
    :param keywords: keyword arguments which are matched to parameter.
    Parameter keywords may be passed using direct format: '-p' or '--param',
    as well as by name: 'p' or 'param', which will be evaluated to '-p' or '--param'.
    Single character parameter will get single hyphen prefix (-p),
    longer parameter names will get double hyphen prefix (--param)
    :param name: explicit internal paramter name (can be used to distinguish it from any keyword)
    :param help: description of the parameter displayed in help output
    :param required: whether parameter is required.
    If it's required but it's not given, the syntax error will be raised.
    :param default: default value for the parameter, if it's not given (and it's not required)
    :param type: type of parameter value (e.g. str, int, float)
    Reference to a parser function may be provided here as well.
    Then parameter value is evaluated by passing the string argument value to that function.
    :param choices: Explicit list of available choices for the parameter value
    or reference to a function which will be invoked to retrieve such possible values list.
    :param strict_choices: whether given arguments should be validated against available choices
    :param multiple: whether parameter is allowed to occur many times.
    Then parameter has list type and stores list of values
    :return: new parameter rule specification
    """
    return ParameterRule(set(keywords), name, type, choices, strict_choices, required, default, help, multiple)


def dictionary(
        *keywords: str,
        name: str = None,
        help: str = None,
        key_type: Union[Type, Callable[[str], Any]] = str,
        value_type: Union[Type, Callable[[str], Any]] = str,
) -> DictionaryRule:
    """
    Create dictionary rule specification.
    Dictionary contains key-value pairs.
    You can add multiple values to it by passing arguments in a manner:
    '-c name1 value1 -c name2 value2'.
    By default it stores empty Python dict.
    These values may be later referenced as dict by its explicit name or keywords
    (in lowercase format without hyphen prefix and with underscores instead of dashes,
    e.g. '--config-name' will be injected as 'config_name')
    :param keywords: keyword arguments which are matched to this dictionary.
    Keywords may be passed using direct format: '-c' or '--config',
    as well as by name: 'c' or 'config', which will be evaluated to '-c' or '--config'.
    Single character dictionary will get single hyphen prefix (-c),
    longer dictionary names will get double hyphen prefix (--config)
    :param name: explicit internal dictionary name (can be used to distinguish it from any keyword)
    :param help: description of the dictionary displayed in help output
    :param key_type: type of dictionary key (e.g. str, int, float)
    Reference to a parser function may be provided here as well.
    Then dictionary value is evaluated by passing the string argument value to that function.
    :param value_type: type of dictionary value (e.g. str, int, float)
    Reference to a parser function may be provided here as well.
    Then dictionary value is evaluated by passing the string argument value to that function.
    :return: new dictionary rule specification
    """
    return DictionaryRule(set(keywords), help, name, key_type, value_type)


def argument(
        name: str,
        help: str = None,
        required: bool = True,
        default: Any = None,
        type: Union[Type, Callable[[str], Any]] = str,
        choices: Union[Iterable[Any], Callable[..., List[Any]]] = None,
        strict_choices: bool = False,
) -> PositionalArgumentRule:
    """
    Create positional argument rule specification.
    Positional argument is an unnamed param
    which is recognized by its position in the command line arguments list.
    :param name: internal argument name, which will be used to reference argument value
    :param help: description of the argument displayed in help output
    :param required: whether positional argument is required.
    If it's required but it's not given, the syntax error will be raised.
    :param default: default value for the argument, if it's not given (and it's not required)
    :param type: explicit type of argument value (e.g. str, int, float)
    Reference to a parser function may be provided here as well.
    Then argument value is evaluated by passing the string argument value to that function.
    :param choices: Explicit list of available choices for the argument value
    or reference to a function which will be invoked to retrieve such possible values list.
    :param strict_choices: whether given arguments should be validated against available choices
    :return: new positional argument rule specification
    """
    return PositionalArgumentRule(name, type, choices, strict_choices, required, default, help)


def arguments(
        name: str,
        type: Union[Type, Callable[[str], Any]] = str,
        choices: Union[Iterable[Any], Callable[..., List[Any]]] = None,
        strict_choices: bool = False,
        count: Optional[int] = None,
        min_count: Optional[int] = None,
        max_count: Optional[int] = None,
        joined_with: Optional[str] = None,
        help: str = None,
) -> ManyArgumentsRule:
    """
    Create 'Multiple arguments' rule specification.
    It allows to retrieve specific number of CLI argumetns or all remaining arguments.
    All matched arguments will be extracted to a list of arguments or a string (depending on joined_with parameter)
    :param name: internal variable name, which will be used to reference matched arguments
    :param type: explicit type of arguments values (e.g. str, int, float)
    Reference to a parser function may be provided here as well.
    Then argument value is evaluated by passing the string argument value to that function.
    :param choices: Explicit list of available choices for the argument value
    or reference to a function which will be invoked to retrieve such possible values list.
    :param strict_choices: whether given arguments should be validated against available choices
    :param count: explicit number of arguments to retrieve.
    If undefined, there is no validation for arguments count.
    If you need particular number of arguments, you can use this count instead of setting min_count=max_count.
    :param min_count: minimum number of arguments.
    By default, there is no lower limit (it is 0).
    :param max_count: maximum number of arguments.
    If undefined, there is no upper limit for arguments count.
    :param joined_with: optional string joiner for arguments.
    If it's set, all matched arguments will be joined to string with that joiner.
    It it's not given, matched arguments will be passed as list of strings.
    This value (string or list) can be accessed by specified name, when it's being injected to a function.
    :param help: description of the arguments displayed in help output
    :return: new many arguments rule specification
    """
    return ManyArgumentsRule(name, type, choices, strict_choices, help, count, min_count, max_count, joined_with)


def default_action(
        run: Callable[..., None] = None,
) -> DefaultActionRule:
    """
    Sets default action for CLI Builder or subcommand.
    Default action is a function which is invoked when no other, more specific rule is matched.
    Any parameters, flags and named arguments which are parameters of action function
    will be automatically injected based on parsed arguments.
    :param run: reference to a function wchich will be invoked
    :return: new default action specification
    """
    return DefaultActionRule(run)


def primary_option(
        *keywords: str,
        run: Callable[..., None] = None,
        help: str = None,
) -> PrimaryOptionRule:
    """
    Create Primary option rule specification.
    Primary option is a keyword which instantly triggers an action when it's detected,
    e.g. '--help', '--version'. It runs only specified action and stops further analyzing.
    :param keywords: keyword arguments, any of them triggers action from primary option
    :param run: action to be invoked when primary option is matched
    :param help: description of the primary option displayed in help output
    :return: new primary option rule specification
    """
    return PrimaryOptionRule(help, set(keywords), run)



from nuclear import *
from tests.asserts import MockIO
from tests.parser.actions import *


def test_setting_flag():
    with MockIO('--force', 'run') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_flag_force),
            flag('--force'),
        ).run()
        assert mockio.output() == 'force: True\n'
    with MockIO('run', '-f') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_flag_force),
            flag('--force', '-f'),
        ).run()
        assert mockio.output() == 'force: True\n'


def test_only_flag():
    with MockIO('--force') as mockio:
        CliBuilder(run=print_flag_force).has(
            flag('--force', '-f'),
        ).run()
        assert mockio.output() == 'force: True\n'


def test_no_flag_set():
    with MockIO() as mockio:
        CliBuilder(run=print_flag_force).has(
            flag('--force', '-f'),
        ).run()
        assert mockio.output() == 'force: False\n'


def test_no_flag_defined():
    with MockIO('--force') as mockio:
        CliBuilder(run=print_flag_force, error_unrecognized=False).run()
        assert 'force: None' in mockio.output()
    with MockIO('--force') as mockio:
        CliBuilder(run=print_flag_force, error_unrecognized=True).run()
        assert 'unrecognized arguments' in mockio.output()


def test_adding_flag_by_name():
    with MockIO('--force') as mockio:
        CliBuilder(run=print_flag_force).has(
            flag('force'),
        ).run()
        assert mockio.output() == 'force: True\n'


def test_adding_flag_by_short_name():
    def print_flag_f(f: bool):
        print(f'force: {f}')

    with MockIO('-f') as mockio:
        CliBuilder(run=print_flag_f).has(
            flag('f'),
        ).run()
        assert mockio.output() == 'force: True\n'


def test_dashed_param_name():
    def print_flag(skip_it: bool):
        print(f'skip_it: {skip_it}')

    with MockIO('--skip-it') as mockio:
        CliBuilder(run=print_flag).has(
            flag('skip-it'),
        ).run()
        assert mockio.output() == 'skip_it: True\n'


def test_multi_flag():
    def print_flag(verbose: int, v: int):
        print(f'verbose: {verbose} {v}')

    with MockIO('-v', '--verbose', '-v') as mockio:
        CliBuilder(run=print_flag).has(
            flag('verbose', 'v', multiple=True),
        ).run()
        assert mockio.output() == 'verbose: 3 3\n'

    with MockIO() as mockio:
        CliBuilder(run=print_flag).has(
            flag('verbose', 'v', multiple=True),
        ).run()
        assert mockio.output() == 'verbose: 0 0\n'


def test_combined_short_single_flags():
    def print_flags(t: bool, u: bool, l: bool):
        print(f'flags: {t} {u} {l}')

    with MockIO('-tul') as mockio:
        CliBuilder(run=print_flags).has(
            flag('t'),
            flag('u'),
            flag('l'),
        ).run()
        assert mockio.stripped() == 'flags: True True True'


def test_combined_short_multi_flag():
    def print_flags(verbose: int):
        print(f'verbose level: {verbose}')

    with MockIO('-vvv') as mockio:
        CliBuilder(run=print_flags).has(
            flag('verbose', 'v', multiple=True),
        ).run()
        assert mockio.stripped() == 'verbose level: 3'

    with MockIO() as mockio:
        CliBuilder(run=print_flags).has(
            flag('verbose', 'v', multiple=True),
        ).run()
        assert mockio.stripped() == 'verbose level: 0'



from pathlib import Path

from nuclear import *
from nuclear.completers import file_completer
from tests.asserts import MockIO


def test_file_completer_empty():
    with MockIO('--autocomplete', '"app "') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        proposals = mockio.stripped().splitlines()
        assert '.gitignore' in proposals
        assert 'tests' in proposals

        assert 'tests/' not in proposals
        assert 'tests/autocomplete' not in proposals
        assert 'tests/__init__.py' not in proposals
        assert '.' not in proposals
        assert '..' not in proposals


def test_file_completer_file():
    with MockIO('--autocomplete', '"app .gitignore"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        assert mockio.stripped() == '.gitignore'


def test_file_completer_dir1():
    with MockIO('--autocomplete', '"app tests"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        proposals = set(mockio.stripped().splitlines())
        assert proposals == {'tests'}


def test_file_completer_dir_content():
    with MockIO('--autocomplete', '"app tests/"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        proposals = mockio.stripped().splitlines()
        assert 'tests/autocomplete' in proposals
        assert 'tests/__init__.py' in proposals

        assert 'tests/' not in proposals
        assert 'tests/autocomplete/' not in proposals
        assert '.gitignore' not in proposals
        assert '.' not in proposals
        assert '..' not in proposals


def test_file_completer_subdir():
    with MockIO('--autocomplete', '"app tests/autocomplete"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        proposals = set(mockio.stripped().splitlines())
        assert proposals == {'tests/autocomplete'}


def test_file_completer_subdir_content():
    with MockIO('--autocomplete', '"app tests/autocomplete/"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        proposals = mockio.stripped().splitlines()
        assert 'tests/autocomplete/__init__.py' in proposals

        assert 'tests/autocomplete/' not in proposals
        assert 'tests/autocomplete' not in proposals
        assert 'tests/__init__.py' not in proposals
        assert 'tests/' not in proposals
        assert '.gitignore' not in proposals
        assert '.' not in proposals
        assert '..' not in proposals


def test_file_completer_subdir_file():
    with MockIO('--autocomplete', '"app tests/autocomplete/__init__.py"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        assert mockio.stripped() == 'tests/autocomplete/__init__.py'


def test_file_completer_notexisting_dir():
    with MockIO('--autocomplete', '"app tests/there_is_no_dir/__init__.py"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        assert mockio.stripped() == ''


def test_complete_absolute_files():
    Path('/tmp/nuclear_test_autocomplete_425_896').write_text('')

    with MockIO('--autocomplete', '"app /tmp/nuclear_test_autocomplete_425"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('f', choices=file_completer),
        ).run()
        assert mockio.stripped() == '/tmp/nuclear_test_autocomplete_425_896'

    Path('/tmp/nuclear_test_autocomplete_425_896').unlink()



import os
import sys
from dataclasses import dataclass, field
from typing import List, Set, Optional

from nuclear.builder.rule import PrimaryOptionRule, ParameterRule, FlagRule, CliRule, SubcommandRule, \
    PositionalArgumentRule, ManyArgumentsRule, DictionaryRule, ValueRule
from nuclear.parser.context import RunContext
from nuclear.parser.keyword import format_var_names, format_var_name
from nuclear.parser.parser import Parser
from nuclear.parser.transform import filter_rules
from nuclear.parser.value import generate_value_choices
from nuclear.version import __version__


@dataclass
class _OptionHelp(object):
    cmd: str
    help: str
    parent: '_OptionHelp' = None
    rule: SubcommandRule = None
    subrules: List[CliRule] = field(default_factory=lambda: [])


internal_options = {'--autocomplete', '--install-bash', '--install-autocomplete'}


def print_help(rules: List[CliRule], app_name: str, version: str, help: str, subargs: List[str], hide_internal: bool):
    helps = generate_help(rules, app_name, version, help, subargs, hide_internal)
    print('\n'.join(helps))


def print_usage(rules: List[CliRule]):
    all_rules, available_subcommands, precommands = help_context(rules, [])

    pos_arguments = filter_rules(all_rules, PositionalArgumentRule)
    many_args = filter_rules(all_rules, ManyArgumentsRule)

    has_commands = bool(filter_rules(available_subcommands, SubcommandRule))

    command_name = shell_command_name()
    app_bin_prefix = ' '.join([command_name] + precommands)

    usage = generate_usage(app_bin_prefix, has_commands, have_rules_options(all_rules), many_args, pos_arguments)

    how_to_help = f'Run "{command_name} --help" for more information.'
    print('\n'.join([f'Usage: {usage}', how_to_help]))


def generate_help(rules: List[CliRule], app_name: str, version: str, help: str, subargs: List[str],
                  hide_internal: bool) -> List[str]:
    all_rules, available_subcommands, precommands = help_context(rules, subargs)
    return generate_subcommand_help(all_rules, app_name, version, help,
                                    precommands, available_subcommands, hide_internal)


def help_context(rules, subargs):
    available_subcommands = filter_rules(rules, SubcommandRule)
    run_context: Optional[RunContext] = Parser(rules, dry=True).parse_args(subargs)
    all_rules: List[CliRule] = run_context.active_rules
    active_subcommands: List[SubcommandRule] = run_context.active_subcommands
    precommands: List[str] = [_subcommand_short_name(rule) for rule in active_subcommands]
    if active_subcommands:
        available_subcommands = filter_rules(active_subcommands[-1].subrules, SubcommandRule)
    return all_rules, available_subcommands, precommands


def generate_subcommand_help(
        all_rules: List[CliRule],
        app_name: str,
        version: str,
        help: str,
        precommands: List[str],
        subcommands: List[SubcommandRule],
        hide_internal: bool,
) -> List[str]:
    pos_arguments = filter_rules(all_rules, PositionalArgumentRule)
    many_args = filter_rules(all_rules, ManyArgumentsRule)

    pos_args_helps: List[_OptionHelp] = _generate_pos_args_helps(pos_arguments, many_args)
    options: List[_OptionHelp] = _generate_options_helps(all_rules, hide_internal)
    commands: List[_OptionHelp] = _generate_commands_helps(subcommands)

    out = []

    app_info = app_help_info(app_name, help, version)
    if app_info:
        out.append(app_info + '\n')

    app_bin_prefix = ' '.join([shell_command_name()] + precommands)
    out.append('Usage:')
    out.append(generate_usage(app_bin_prefix, bool(commands), have_rules_options(all_rules), many_args, pos_arguments))

    if pos_args_helps:
        out.append('\nArguments:')
        __helpers_output(pos_args_helps, out)

    if options:
        out.append('\nOptions:')
        __helpers_output(options, out)

    if commands:
        out.append('\nCommands:')
        __helpers_output(commands, out)
        out.append(f'\nRun "{app_bin_prefix} COMMAND --help" for more information on a command.')

    return out


def app_help_info(app_name: str, help: str, version: str) -> Optional[str]:
    info = app_name_version(app_name, version)
    return ' - '.join(filter(bool, [info, help]))


def app_name_version(app_name, version):
    infos = []
    if app_name:
        infos += [app_name]
    if version:
        version = _normalized_version(version)
        infos += [version]
    if infos:
        infos += [f'(nuclear v{__version__})']
    return ' '.join(infos)


def generate_usage(app_bin_prefix, has_commands: bool, has_options: bool, many_args, pos_arguments) -> str:
    usage_syntax: str = app_bin_prefix
    if has_commands:
        usage_syntax += ' [COMMAND]'
    if has_options:
        usage_syntax += ' [OPTIONS]'
    usage_syntax += usage_positional_arguments(pos_arguments)
    usage_syntax += usage_many_arguments(many_args)
    return usage_syntax


def __helpers_output(commands, out):
    padding = _max_name_width(commands)
    for helper in commands:
        name_padded = helper.cmd.ljust(padding)
        if helper.help:
            for idx, line in enumerate(helper.help.splitlines()):
                if idx == 0:
                    out.append(f'  {name_padded} - {line}')
                else:
                    out.append(' ' * (2 + padding + 3) + line)
        else:
            out.append(f'  {name_padded}')


def print_version(app_name: str, version: str):
    print(app_name_version(app_name, version))


def _normalized_version(version: str) -> str:
    if version.startswith('v'):
        return version
    return f'v{version}'


def _max_name_width(helps: List[_OptionHelp]) -> int:
    return max(map(lambda h: len(h.cmd), helps))


def _generate_pos_args_helps(
        pos_arguments: List[PositionalArgumentRule],
        many_args: List[ManyArgumentsRule]
) -> List[_OptionHelp]:
    return [_pos_arg_help(rule) for rule in pos_arguments] + \
           [_many_args_help(rule) for rule in many_args]


def _generate_options_helps(rules: List[CliRule], hide_internal: bool) -> List[_OptionHelp]:
    # filter non-empty
    return list(filter(lambda o: o, [_generate_option_help(rule, hide_internal) for rule in rules]))


def _generate_option_help(rule: CliRule, hide_internal: bool) -> Optional[_OptionHelp]:
    if isinstance(rule, PrimaryOptionRule):
        return _primary_option_help(rule, hide_internal)
    elif isinstance(rule, FlagRule):
        return _flag_help(rule)
    elif isinstance(rule, ParameterRule):
        return _parameter_help(rule)
    elif isinstance(rule, DictionaryRule):
        return _dictionary_help(rule)
    return None


def _generate_commands_helps(rules: List[CliRule], parent: _OptionHelp = None, subrules: List[CliRule] = None
                             ) -> List[_OptionHelp]:
    commands: List[_OptionHelp] = []
    for rule in filter_rules(rules, SubcommandRule):
        subsubrules = (subrules or []) + rule.subrules
        helper = _subcommand_help(rule, parent, subsubrules)
        if rule.run or rule.help:
            commands.append(helper)
        commands.extend(_generate_commands_helps(rule.subrules, helper, subsubrules))
    return commands


def _subcommand_help(rule: SubcommandRule, parent: _OptionHelp, subrules: List[CliRule]) -> _OptionHelp:
    pos_args = filter_rules(subrules, PositionalArgumentRule)
    many_args = filter_rules(subrules, ManyArgumentsRule)
    cmd = _subcommand_prefix(parent) + '|'.join(sorted_keywords(rule.keywords))
    cmd += usage_positional_arguments(pos_args)
    cmd += usage_many_arguments(many_args)
    return _OptionHelp(cmd, rule.help, parent=parent, rule=rule, subrules=subrules)


def _subcommand_prefix(helper: _OptionHelp) -> str:
    if not helper:
        return ''
    return _subcommand_prefix(helper.parent) + '|'.join(sorted_keywords(helper.rule.keywords)) + ' '


def _primary_option_help(rule: PrimaryOptionRule, hide_internal: bool) -> Optional[_OptionHelp]:
    if hide_internal:
        for keyword in rule.keywords:
            if keyword in internal_options:
                return None
    cmd = ', '.join(sorted_keywords(rule.keywords))
    pos_args = filter_rules(rule.subrules, PositionalArgumentRule)
    all_args = filter_rules(rule.subrules, ManyArgumentsRule)
    cmd += usage_positional_arguments(pos_args)
    cmd += usage_many_arguments(all_args)
    return _OptionHelp(cmd, rule.help)


def _flag_help(rule: FlagRule) -> _OptionHelp:
    cmd = ', '.join(sorted_keywords(rule.keywords))
    return _OptionHelp(cmd, rule.help)


def _parameter_help(rule: ParameterRule) -> _OptionHelp:
    cmd = ', '.join(sorted_keywords(rule.keywords)) + ' ' + _param_display_name(rule)
    default_value = display_default_value(rule.default)
    choices_help = display_choices_help(rule)
    help_text = join_nonempty_lines(rule.help, default_value, choices_help)
    return _OptionHelp(cmd, help_text)


def _dictionary_help(rule: DictionaryRule) -> _OptionHelp:
    cmd = ', '.join(sorted_keywords(rule.keywords)) + ' KEY VALUE'
    return _OptionHelp(cmd, rule.help)


def _pos_arg_help(rule: PositionalArgumentRule) -> _OptionHelp:
    cmd = display_positional_argument(rule)
    default_value = display_default_value(rule.default)
    choices_help = display_choices_help(rule)
    help_text = join_nonempty_lines(rule.help, default_value, choices_help)
    return _OptionHelp(cmd, help_text)


def _many_args_help(rule: ManyArgumentsRule) -> _OptionHelp:
    cmd = display_many_arguments(rule)
    choices_help = display_choices_help(rule)
    help_text = join_nonempty_lines(rule.help, choices_help)
    return _OptionHelp(cmd, help_text)


def _param_display_name(rule: ParameterRule) -> str:
    if rule.name:
        return format_var_name(rule.name).upper()
    else:
        # get name from longest keyword
        names: Set[str] = format_var_names(rule.keywords)
        return max(names, key=lambda n: len(n)).upper()


def _argument_var_name(rule: PositionalArgumentRule) -> str:
    return format_var_name(rule.name).upper()


def _subcommand_short_name(rule: SubcommandRule) -> str:
    return next(iter(rule.keywords))


def sorted_keywords(keywords: Set[str]) -> List[str]:
    # shortest keywords first, then alphabetically
    return sorted(keywords, key=lambda k: (len(k), k))


def display_positional_argument(rule: PositionalArgumentRule) -> str:
    var_name = _argument_var_name(rule)
    if rule.required:
        return f' {var_name}'
    else:
        return f' [{var_name}]'


def display_many_arguments(rule: ManyArgumentsRule) -> str:
    arg_name = rule.name.upper()
    if rule.count_min():
        return f' {arg_name}...'
    else:
        return f' [{arg_name}...]'


def usage_positional_arguments(rules: List[PositionalArgumentRule]) -> str:
    return ''.join([display_positional_argument(rule) for rule in rules])


def usage_many_arguments(rules: List[ManyArgumentsRule]) -> str:
    return ''.join([display_many_arguments(rule) for rule in rules])


def shell_command_name():
    _, command = os.path.split(sys.argv[0])
    return command


def have_rules_options(rules: List[CliRule]) -> bool:
    return bool(filter_rules(rules, FlagRule, ParameterRule, DictionaryRule, PrimaryOptionRule))


def display_default_value(default) -> Optional[str]:
    if default is None:
        return None
    return 'Default: ' + str(default)


def display_choices_help(rule: ValueRule) -> Optional[str]:
    choices = generate_value_choices(rule)
    if not choices or not rule.strict_choices:
        return None
    return 'Choices: ' + ', '.join(choices)


def join_nonempty_lines(*lines: str) -> str:
    return '\n'.join(filter(lambda t: t is not None, lines))



from nuclear import parameter
from nuclear.parser.value import generate_value_choices


def test_generate_choices_from_list():
    assert generate_value_choices(parameter('key', choices=[])) == []
    assert generate_value_choices(parameter('key', choices=['abc'])) == ['abc']
    assert generate_value_choices(parameter('key', choices=['1', '2', '3'])) == ['1', '2', '3']


def test_generate_choices_from_set():
    assert generate_value_choices(parameter('key', choices=set())) == []
    assert generate_value_choices(parameter('key', choices={'abc'})) == ['abc']


def test_generate_choices_from_tuple():
    assert generate_value_choices(parameter('key', choices=())) == []
    assert generate_value_choices(parameter('key', choices=('abc',))) == ['abc']
    assert generate_value_choices(parameter('key', choices=('1', '2', '3'))) == ['1', '2', '3']


def test_generate_choices_from_dicts():
    items = {'a': '1'}
    assert generate_value_choices(parameter('key', choices=items.keys())) == ['a']
    assert generate_value_choices(parameter('key', choices=items.values())) == ['1']


def test_generate_choices_from_callable():
    def provide():
        return ['val', '2']

    assert generate_value_choices(parameter('key', choices=provide)) == ['val', '2']



class CliError(RuntimeError):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.__msg = msg

    def __str__(self):
        if self.__cause__:
            return f'{self.__msg}: {self.__cause__}'
        return self.__msg


class CliDefinitionError(CliError):
    def __init__(self, msg: str):
        super().__init__(msg)


class CliSyntaxError(CliError):
    def __init__(self, msg: str):
        super().__init__(msg)



from typing import Optional, Tuple

from nuclear.args.args_que import ArgsQue
from nuclear.builder.rule import DictionaryRule
from nuclear.builder.rule import ParameterRule
from .error import CliSyntaxError


def match_param(rule: ParameterRule, args: ArgsQue, arg: str) -> Optional[str]:
    for keyword in rule.keywords:
        # match 2 args: --name value
        if arg == keyword:
            args.pop_current()
            if not args.has_next():
                raise CliSyntaxError('missing value argument for parameter')
            return args.pop_current()
        # match 1 arg: --name=value
        prefix = keyword + '='
        if arg.startswith(prefix):
            return args.pop_current()[len(prefix):]
    return None


def match_dictionary(rule: DictionaryRule, args: ArgsQue, arg: str
                     ) -> Tuple[Optional[str], Optional[str]]:
    for keyword in rule.keywords:
        # match 3 args: --keyword name value
        if arg == keyword:
            args.pop_current()
            if not args.has_next():
                raise CliSyntaxError('missing key argument for dictionary')
            var_key = args.pop_current()
            if not args.has_next():
                raise CliSyntaxError('missing value argument for dictionary')
            var_value = args.pop_current()
            return var_key, var_value
    return None, None



from nuclear import *
from tests.asserts import MockIO


def list_screens():
    return ['HDMI', 'eDP']


def cli_subparameter():
    return CliBuilder().has(
        subcommand('sub').has(
            parameter('--parameter', choices=list_screens),
        ),
    )


def test_no_match():
    with MockIO('--autocomplete', 'app nomatch') as mockio:
        CliBuilder().run()
        assert mockio.output() == '\n'


def test_empty_builder_proposals():
    with MockIO('--autocomplete', 'app ') as mockio:
        CliBuilder(with_defaults=True, version='1').run()
        assert '-h\n' in mockio.output()
        assert '--help\n' in mockio.output()
        assert '--version\n' in mockio.output()
        assert '--install-bash\n' in mockio.output()
        assert '--autocomplete\n' in mockio.output()
    with MockIO('--autocomplete', '"nomatch "') as mockio:
        CliBuilder(with_defaults=True).run()
        assert '--help\n' in mockio.output()


def cli_subcommands():
    return CliBuilder().has(
        subcommand('git').has(
            subcommand('push'),
            subcommand('remote').has(
                subcommand('set-url', 'rename'),
            ),
        ),
        subcommand('xrandr'),
    )


def test_autocomplete_base_command():
    with MockIO('--autocomplete', 'app ') as mockio:
        cli_subcommands().run()
        assert 'git\n' in mockio.output()
        assert 'xrandr\n' in mockio.output()


def test_skip_nomatching_word():
    with MockIO('--autocomplete', '"app nomatch "') as mockio:
        cli_subcommands().run()
        assert 'git\n' in mockio.output()
        assert 'xrandr\n' in mockio.output()


def test_autocomplete_subcommands():
    with MockIO('--autocomplete', '"app g"') as mockio:
        cli_subcommands().run()
        assert mockio.output() == 'git\n'
    with MockIO('--autocomplete', '"app git"') as mockio:
        cli_subcommands().run()
        assert mockio.output() == 'git\n'
    with MockIO('--autocomplete', '"app git "') as mockio:
        cli_subcommands().run()
        assert 'push\n' in mockio.output()
        assert 'remote\n' in mockio.output()
        assert 'git\n' not in mockio.output()
        assert 'xrandr\n' not in mockio.output()
    with MockIO('--autocomplete', '"app git r"') as mockio:
        cli_subcommands().run()
        assert 'remote\n' == mockio.output()
    with MockIO('--autocomplete', '"app git remote"') as mockio:
        cli_subcommands().run()
        assert 'remote\n' == mockio.output()


def test_autocomplete_3rd_level_subcommand():
    with MockIO('--autocomplete', '"app git remote "') as mockio:
        cli_subcommands().run()
        assert 'set-url\n' in mockio.output()
        assert 'remote\n' not in mockio.output()
        assert 'git\n' not in mockio.output()


def test_autocomplete_flags():
    def cli_flags():
        return CliBuilder().has(
            subcommand('sub').has(
                flag('--flag-local'),
            ),
            flag('--flag-global'),
        )

    with MockIO('--autocomplete', 'app') as mockio:
        cli_flags().run()
        assert '--flag-global\n' in mockio.output()
        assert '--flag-local\n' not in mockio.output()
    with MockIO('--autocomplete', 'app sub ') as mockio:
        cli_flags().run()
        assert '--flag-global\n' in mockio.output()
        assert '--flag-local\n' in mockio.output()


def test_autocomplete_parameters():
    def cli_parameters():
        return CliBuilder().has(
            subcommand('sub').has(
                parameter('--param-local'),
            ),
            parameter('--param-global'),
        )

    with MockIO('--autocomplete', 'app') as mockio:
        cli_parameters().run()
        assert '--param-global\n' in mockio.output()
        assert '--param-local\n' not in mockio.output()
    with MockIO('--autocomplete', 'app sub ') as mockio:
        cli_parameters().run()
        assert '--param-global\n' in mockio.output()
        assert '--param-local\n' in mockio.output()


def test_autocomplete_parameter_name():
    with MockIO('--autocomplete', '"app sub --param"') as mockio:
        cli_subparameter().run()
        assert '--parameter\n' in mockio.output()
        assert '--parameter=\n' in mockio.output()


def test_autocomplete_parameters_choice_2params():
    with MockIO('--autocomplete', '"app sub --parameter"') as mockio:
        cli_subparameter().run()
        assert '--parameter\n' in mockio.output()
        assert '--parameter=\n' in mockio.output()
    with MockIO('--autocomplete', '"app sub --parameter "') as mockio:
        cli_subparameter().run()
        assert 'HDMI\n' in mockio.output()
        assert 'eDP\n' in mockio.output()
    with MockIO('--autocomplete', '"app sub --parameter H"') as mockio:
        cli_subparameter().run()
        assert mockio.output() == 'HDMI\n'


def test_autocomplete_parameters_choice_equal():
    with MockIO('--autocomplete', '"app sub --parameter="') as mockio:
        cli_subparameter().run()
        assert '--parameter=HDMI\n' not in mockio.output()
        assert 'HDMI\n' in mockio.output()
        assert 'eDP\n' in mockio.output()
    with MockIO('--autocomplete', '"app sub --parameter=H"') as mockio:
        cli_subparameter().run()
        assert mockio.output() == 'HDMI\n'


def test_autocomplete_pos_arguments_choice():
    def cli_argument():
        return CliBuilder().has(
            subcommand('sub').has(
                argument('pos', choices=['abc', 'def']),
            ),
        )

    with MockIO('--autocomplete', '"app sub pos "') as mockio:
        cli_argument().run()
        assert 'abc\n' in mockio.output()
        assert 'def' in mockio.output()
    with MockIO('--autocomplete', '"app sub pos a"') as mockio:
        cli_argument().run()
        assert 'abc\n' in mockio.output()
        assert 'def' not in mockio.output()


def test_autocomplete_many_arguments_choice():
    def cli_argument():
        return CliBuilder().has(
            arguments('words', choices=['abc', 'def']),
        )

    with MockIO('--autocomplete', '"app "') as mockio:
        cli_argument().run()
        assert 'abc\n' in mockio.output()
        assert 'def' in mockio.output()


def test_autocomplete_command_last_word_space():
    with MockIO('--autocomplete', 'app info age') as mockio:
        CliBuilder().has(
            subcommand('info').has(
                subcommand('age'),
            ),
        ).run()
        assert 'age\n' in mockio.output()


def test_doubled_proposals():
    with MockIO('--autocomplete', 'app --version ') as mockio:
        CliBuilder(version='1').run()
        assert '-h\n' in mockio.output()
        assert '--help\n' in mockio.output()
        assert '--version\n' in mockio.output()
        assert '--install-bash\n' in mockio.output()
        assert '--autocomplete\n' in mockio.output()


def test_empty_choices():
    with MockIO('--autocomplete', '"app sub --pos "') as mockio:
        CliBuilder().has(
            parameter('pos', choices=[]),
        ).run()
        assert mockio.stripped() == ''


def test_opened_quote():
    with MockIO('--autocomplete', '"app \""') as mockio:
        CliBuilder().has(
            parameter('pos', choices=['val']),
        ).run()
        assert mockio.stripped() == ''


def test_completing_word_in_the_middle():
    with MockIO('--autocomplete', '"app ru in"', '1') as mockio:
        CliBuilder().has(
            subcommand('info'),
            subcommand('run'),
        ).run()
        assert mockio.stripped() == 'run'


def test_complete_with_completer_function():
    def complete():
        return ['42', '47', '53']

    with MockIO('--autocomplete', '"app 4"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('a', choices=complete),
        ).run()
        assert mockio.stripped() == '42\n47'


def test_escaping_space_filenames():
    def complete():
        return ['file with spaces', 'file_without']

    with MockIO('--autocomplete', '"app file"') as mockio:
        CliBuilder(reraise_error=True).has(
            arguments('a', choices=complete),
        ).run()
        assert mockio.stripped() == 'file\\ with\\ spaces\nfile_without'



from nuclear import *
from nuclear.parser.error import CliSyntaxError
from tests.asserts import MockIO, assert_error


def build_builder() -> CliBuilder:
    return CliBuilder('helpgen', version='1.0.0', help='tool for generating help').has(
        subcommand('git').has(
            subcommand('help', help='shows help'),
            subcommand('push').has(
                argument('remote'),
                argument('branch', required=False),
            ),
            subcommand('describe').has(
                flag('--tags', help='show tags'),
            ),
            subcommand('checkout').has(
                argument('branch', choices=['1', '2', '3'], type=str),
                flag('force', '-f'),
            ),
            subcommand('remote', help='show remotes list').has(
                subcommand('set-url', 'rename', help='change remote\'s name').has(
                    argument('remote-name', choices=['1', '2', '3'], type=str),
                    argument('new-name'),
                ),
                flag('force', '-f', help='ignore warnings'),
            ),
            parameter('--count', type=int),
            primary_option('--exit')
        ),
        subcommand('xrandr').has(
            parameter('output', required=True, choices=['HDMI', 'eDP1']),
            flag('--primary'),
        ),
        subcommand('docker').has(
            subcommand('exec').has(
                flag('--it'),
                parameter('-u', name='user'),
                arguments(name='cmd', joined_with=' '),
            ),
            primary_option('--help', '-h'),
        ),
        argument('name'),
        argument('surname', required=False, default=''),
        flag('force'),
    )


def noop():
    pass


def test_default_print_help_empty():
    with MockIO('--help') as mockio:
        CliBuilder(with_defaults=True).run()
        assert 'Usage' in mockio.output()
        assert 'glue [OPTIONS]' in mockio.output()


def test_root_help():
    with MockIO('--help') as mockio:
        build_builder().run()
        assert 'helpgen v1.0.0' in mockio.output()
        assert 'Usage:' in mockio.output()
        assert 'Options:' in mockio.output()
        assert 'Commands:' in mockio.output()
        assert 'Commands:' in mockio.output()

        assert '-h, --help' in mockio.output()
        assert 'Display this help and exit' in mockio.output()
        assert '--version' in mockio.output()

        assert 'git remote rename|set-url' in mockio.output()
        assert 'change remote\'s name' in mockio.output()


def test_subcommand_help():
    with MockIO('git', '--help') as mockio:
        build_builder().run()
        assert 'Usage:' in mockio.output()
        assert 'git [COMMAND]' in mockio.output()
        assert 'xrandr' not in mockio.output()


def test_3rd_level_help():
    with MockIO('git', 'push', '--help') as mockio:
        CliBuilder().has(
            subcommand('git', run=noop).has(
                subcommand('push', run=noop).has(
                    subcommand('remote', run=noop),
                ),
                subcommand('bad', run=noop)
            ),
            subcommand('bad', run=noop)
        ).run()
        assert 'Usage:' in mockio.output()
        assert 'git push [COMMAND]' in mockio.output()
        assert 'remote' in mockio.output()
        assert 'bad' not in mockio.output()


def test_print_help_on_syntax_error():
    with MockIO('--param') as mockio:
        cli = CliBuilder(usage_onerror=True, reraise_error=True).has(
            parameter('param'),
        )
        assert_error(lambda: cli.run(), error_type=CliSyntaxError)
        assert 'Usage:' in mockio.output()


def test_default_help_when_no_arguments():
    with MockIO('') as mockio:
        CliBuilder().run()
        assert 'Usage:' in mockio.output()


def test_hiding_internal_options():
    with MockIO('--help') as mockio:
        CliBuilder(hide_internal=True).run()
        assert '--install-bash' not in mockio.output()
        assert '--autocomplete' not in mockio.output()
    with MockIO('--help') as mockio:
        CliBuilder(hide_internal=False).run()
        assert '--install-bash' in mockio.output()
        assert '--autocomplete' in mockio.output()


def test_omit_empty_subcommands():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            subcommand('show-help', help='shows help'),
            subcommand('push', run=lambda: print('push')),
            subcommand('noop'),
            subcommand('dooo').has(
                subcommand('nothing'),
            ),
            subcommand('do2').has(
                subcommand('it', run=lambda: print('do2 it')),
            ),
        ).run()
        assert 'show-help' in mockio.output()
        assert 'push' in mockio.output()
        assert 'noop' not in mockio.output()
        assert 'dooo' not in mockio.output()
        assert 'nothing' not in mockio.output()
        assert 'do2  ' not in mockio.output()
        assert 'do2 it' in mockio.output()


def test_display_dictionary():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            dictionary('-c', '--config'),
        ).run()
        assert '--config' in mockio.output()


def test_usage_on_syntax_error():
    with MockIO('') as mockio:
        CliBuilder(usage_onerror=True, reraise_error=False).has(
            argument('req')
        ).run()
        assert 'Syntax error' in mockio.output()
        assert '--help" for more information' in mockio.output()


def test_display_explicit_param_name():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            parameter('p', name='param'),
        ).run()
        assert '-p PARAM' in mockio.output()


def test_display_parameter_default_value():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            parameter('src-path', help='source path', default='/home/user'),
            parameter('target-path', default='/root'),
            parameter('num', help='number', type=int, default=52),
        ).run()
        assert 'SRC_PATH' in mockio.output()
        assert 'source path' in mockio.output()
        assert 'Default: /home/user' in mockio.output()
        assert 'Default: 52' in mockio.output()


def test_display_argument_default_value():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            argument('src-path', help='source path', default='/home/user', required=False),
            argument('target-path', default='/root', required=False),
            argument('num', help='number', type=int, default=52, required=False),
        ).run()
        assert 'SRC_PATH' in mockio.output()
        assert 'source path' in mockio.output()
        assert 'Default: /home/user' in mockio.output()
        assert 'Default: 52' in mockio.output()


def test_display_arguments_help():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            argument('src-path', help='source path'),
            argument('num', help='number', type=int),
            arguments('paths'),
        ).run()
        assert 'Arguments:' in mockio.output()
        assert 'SRC_PATH' in mockio.output()
        assert 'source path' in mockio.output()
        assert 'NUM' in mockio.output()
        assert 'number' in mockio.output()
        assert 'PATHS' in mockio.output()


def test_display_arguments_with_subcommand_help():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            subcommand('module', help='install modules').has(
                subcommand('list', help='list modules'),
                arguments('modules', help='module names'),
            )
        ).run()
        mockio.assert_match(r'^ +module \[MODULES\.\.\.\] +- install modules$')
        mockio.assert_match(r'^ +module list \[MODULES...\] +- list modules$')


def test_subcommand_arguments_help():
    with MockIO('module', '--help') as mockio:
        CliBuilder().has(
            subcommand('module', help='install modules').has(
                arguments('modules', help='module names'),
            )
        ).run()
        mockio.assert_match(r'^ +\[MODULES...\] +- module names$')


def test_strict_choices_show_list_help():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            arguments('modules', help='module names', choices=['monty', 'python'], strict_choices=True)
        ).run()
        mockio.assert_match(r'Choices: monty, python$')


def test_unstrict_choices_hidden():
    with MockIO('--help') as mockio:
        CliBuilder().has(
            arguments('modules', help='module names', choices=['monty', 'python'], strict_choices=False)
        ).run()
        assert 'Choices:' not in mockio.output()



from nuclear import *
from nuclear.args.container import ArgsContainer
from tests.asserts import MockIO, assert_error


def test_args_container_by_attr():
    def print_args_dict(args):
        print(' '.join([
            args.param, args.p,
            str(args.p2),
            args.multi_ple_words,
            args.nnn,
        ]))

    with MockIO('--param', 'pval', '--multi-ple-words', 'words', '--p2', '--named-param=mmm') as mockio:
        CliBuilder(run=print_args_dict).has(
            parameter('param', '-p'),
            flag('p2'),
            parameter('multi-ple-words'),
            parameter('named-param', name='nnn'),
        ).run()
        assert mockio.stripped() == 'pval pval True words mmm'


def test_args_container_by_dict_name():
    def print_args_dict(args):
        print(' '.join([
            args['param'], args['--param'], args['p'], args['-p'],
            str(args['p2']), str(args['--p2']),
            args['multi-ple-words'], args['multi_ple_words'],
            args['nnn'],
        ]))

    with MockIO('--param', 'pval', '--multi-ple-words', 'words', '--p2', '--named-param=mmm') as mockio:
        CliBuilder(run=print_args_dict).has(
            parameter('param', '-p'),
            flag('p2'),
            parameter('multi-ple-words'),
            parameter('named-param', name='nnn'),
        ).run()
        assert mockio.stripped() == 'pval pval pval pval True True words words mmm'


def test_args_container_error_on_nonexisting():
    def get_notexisting(args):
        assert_error(lambda: print(args['no_attr']), KeyError)
        assert_error(lambda: print(args.no_such_attr), AttributeError)

    with MockIO('-f'):
        CliBuilder(run=get_notexisting).has(
            flag('f'),
        ).run()


def test_shadowed_args_container_var():
    def print_it(args: ArgsContainer):
        print(args.f)

    with MockIO() as mockio:
        CliBuilder(run=print_it).has(
            flag('f'),
        ).run()
        assert mockio.output() == 'False\n'



from nuclear import *
from tests.asserts import MockIO
from nuclear import __version__


def test_print_version():
    with MockIO('--version') as mockio:
        CliBuilder(version='1.2.3').run()
        assert mockio.stripped().startswith('v1.2.3')


def test_print_version_with_v():
    with MockIO('--version') as mockio:
        CliBuilder(version='v1.2.3').run()
        assert mockio.stripped().startswith('v1.2.3')


def test_print_name_with_version():
    with MockIO('--version') as mockio:
        CliBuilder(name='apk', version='v1.2.3').run()
        assert 'apk v1.2.3' in mockio.stripped()


def test_print_nuclear_version():
    with MockIO('--version') as mockio:
        CliBuilder(name='apk', version='1.2.3').run()
        assert f'apk v1.2.3 (nuclear v{__version__})' == mockio.stripped()


def test_print_nuclear_version_without_app_name():
    with MockIO('--version') as mockio:
        CliBuilder(version='1.2.3').run()
        assert f'v1.2.3 (nuclear v{__version__})' == mockio.stripped()



from nuclear.utils.files import save_file
from nuclear.utils.regex import *


def test_regex_replace():
    assert regex_replace('abca', 'a', 'b') == 'bbcb'
    assert regex_replace('abc123a', r'\d', '5') == 'abc555a'


def test_regex_match():
    assert regex_match('ab 12 def 123', r'.*\d{2}')
    assert not regex_match('ab 1 def 1m2', r'.*\d{2}.*')


def test_regex_list():
    # regexReplaceList
    assert regex_replace_list(['a', 'b', '25', 'c'], r'\d+', '7') == ['a', 'b', '7', 'c']
    assert regex_replace_list(['a1', '2b', '3', ''], r'\d', '7') == ['a7', '7b', '7', '']
    assert regex_replace_list(['a1', '2b', '3', ''], r'.*\d.*', '7') == ['7', '7', '7', '']
    # regexFilterLines
    assert regex_filter_list(['a1', '2b', '3', ''], r'.*\d.*') == ['a1', '2b', '3']
    assert regex_filter_list(['a1', '2b', '3', ''], r'dupa') == []
    # regexSearchFile
    assert regex_search_file('tests/utils/res/findme', r'\t*<author>(\w*)</author>', 1) == 'Anonim'
    assert regex_search_file('tests/utils/res/findme', r'\t*<name>(\w*)</name><sur>(\w*)</sur>', 2) == 'Sur'
    # regexReplaceFile
    save_file('tests/utils/res/replaceme', 'dupa\n123')
    assert regex_replace_file('tests/utils/res/replaceme', r'[a-z]+', 'letters') == 'letters\n123'



import re
import shlex
from typing import List, Optional

from nuclear.builder.rule import CliRule, ParameterRule, FlagRule, SubcommandRule, PrimaryOptionRule, \
    PositionalArgumentRule, ManyArgumentsRule
from nuclear.parser.context import RunContext
from nuclear.parser.parser import Parser
from nuclear.parser.transform import filter_rules
from nuclear.parser.value import generate_value_choices


def bash_autocomplete(rules: List[CliRule], cmdline: str, word_idx: Optional[int]):
    filtered = find_matching_completions(cmdline, rules, word_idx)
    print('\n'.join(filtered))


def find_matching_completions(cmdline, rules, word_idx: Optional[int]) -> List[str]:
    extracted_cmdline = _extract_quotes(cmdline)
    try:
        args: List[str] = extract_args(extracted_cmdline)
    except ValueError:
        return []
    current_word: str = get_current_word(args, word_idx)
    available: List[str] = _find_available_completions(rules, args, current_word)
    # convert '--param=value' proposals to 'value'
    hyphen_param_matcher = re.compile(r'-(.+)=(.+)')
    return [
        escape_spaces(hyphen_param_matcher.sub('\\2', c))
        for c in available
        if c.startswith(current_word)
    ]


def escape_spaces(name: str) -> str:
    return name.replace(' ', '\\ ')


def extract_args(extracted_cmdline) -> List[str]:
    args = shlex.split(extracted_cmdline)[1:]
    # restore last whitespace
    if extracted_cmdline.endswith(' ') or extracted_cmdline.endswith('\t'):
        args.append('')
    return args


def _extract_quotes(cmdline) -> str:
    if cmdline.startswith('"') and cmdline.endswith('"'):
        return cmdline[1:-1]
    return cmdline


def get_current_word(args: List[str], word_idx: Optional[int]) -> str:
    if word_idx is None or word_idx - 1 >= len(args):
        return args[-1] if len(args) > 0 else ''
    return args[word_idx - 1]


def _find_available_completions(rules: List[CliRule], args: List[str], current_word: str) -> List[str]:
    subcommands: List[SubcommandRule] = filter_rules(rules, SubcommandRule)
    run_context: Optional[RunContext] = Parser(rules, dry=True).parse_args(args)
    all_rules: List[CliRule] = run_context.active_rules
    active_subcommands: List[SubcommandRule] = run_context.active_subcommands
    if active_subcommands:
        subcommands = filter_rules(active_subcommands[-1].subrules, SubcommandRule)

        # current word is exactly the last command
        if current_word in active_subcommands[-1].keywords:
            return [current_word]

    flags = filter_rules(all_rules, FlagRule)
    parameters = filter_rules(all_rules, ParameterRule)
    primary_options = filter_rules(all_rules, PrimaryOptionRule)
    pos_arguments = filter_rules(all_rules, PositionalArgumentRule)
    many_args = filter_rules(all_rules, ManyArgumentsRule)

    # "--param value" autocompletion
    previous: Optional[str] = args[-2] if len(args) > 1 else None
    if previous:
        for rule in parameters:
            for keyword in rule.keywords:
                if previous == keyword:
                    possible_choices: List[str] = generate_value_choices(rule, current=current_word)
                    return possible_choices

    # "--param=value" autocompletion
    for rule in parameters:
        for keyword in rule.keywords:
            if current_word.startswith(keyword + '='):
                possible_choices: List[str] = list(map(lambda c: keyword + '=' + c,
                                                       generate_value_choices(rule, current=current_word)))
                return possible_choices

    completions: List[str] = []
    # subcommands
    for rule in subcommands:
        completions.extend(rule.keywords)

    # flags, parameter names, primary options
    for rule in flags:
        completions.extend(rule.keywords)

    for rule in parameters:
        for keyword in rule.keywords:
            completions.append(keyword)
            completions.append(keyword + '=')

    for rule in primary_options:
        completions.extend(rule.keywords)

    # positional arguments
    for rule in pos_arguments:
        possible_choices: List[str] = generate_value_choices(rule, current=current_word)
        completions.extend(possible_choices)
    for rule in many_args:
        possible_choices: List[str] = generate_value_choices(rule, current=current_word)
        completions.extend(possible_choices)

    return completions



from typing import List

from nuclear import *
from tests.asserts import MockIO, assert_cli_error
from tests.parser.actions import *


def test_param_with_2_args():
    with MockIO('--param', 'OK') as mockio:
        CliBuilder(run=print_param).has(
            parameter('--param'),
        ).run()
        assert mockio.output() == 'param: OK\n'
    with MockIO('-p', 'OK') as mockio:
        CliBuilder(run=print_param).has(
            parameter('--param', '-p'),
        ).run()
        assert mockio.output() == 'param: OK\n'


def test_param_with_2_args_with_action():
    with MockIO('run', '--param', 'OK') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_param),
            parameter('param'),
        ).run()
        assert mockio.output() == 'param: OK\n'
    with MockIO('run', '-p', 'OK') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_param),
            parameter('param', 'p'),
        ).run()
        assert mockio.output() == 'param: OK\n'


def test_set_param_with_equals():
    with MockIO('run', '--param=OK') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_param),
            parameter('param'),
        ).run()
        assert mockio.output() == 'param: OK\n'
    with MockIO('run', '-p=OK') as mockio:
        CliBuilder().has(
            subcommand('run', run=print_param),
            parameter('param', 'p'),
        ).run()
        assert mockio.output() == 'param: OK\n'
    with MockIO('--param=OK') as mockio:
        CliBuilder(run=print_param).has(
            parameter('--param'),
        ).run()
        assert mockio.output() == 'param: OK\n'
    with MockIO('-p=OK') as mockio:
        CliBuilder(run=print_param).has(
            parameter('--param', '-p'),
        ).run()
        assert mockio.output() == 'param: OK\n'


def test_no_param():
    with MockIO() as mockio:
        CliBuilder(run=print_param).has(
            parameter('--param', '-p'),
        ).run()
        assert mockio.output() == 'param: None\n'


def test_int_type_param():
    def print_param_and_type(param):
        print(f'param: {param}, type: {type(param).__name__}')

    with MockIO('--param', '42') as mockio:
        CliBuilder(run=print_param_and_type).has(
            parameter('--param', type=int),
        ).run()
        assert mockio.output() == 'param: 42, type: int\n'


def test_missing_required_param():
    with MockIO():
        cli = CliBuilder(run=print_param, usage_onerror=False, reraise_error=True).has(
            parameter('--param', required=True),
        )
        assert_cli_error(lambda: cli.run())


def test_multi_parameters():
    def sum_param(param: List[int]):
        print(f'param: {sum(param)}')

    with MockIO('--param', '800', '--param=42') as mockio:
        CliBuilder(run=sum_param).has(
            parameter('param', multiple=True, type=int),
        ).run()
        assert mockio.output() == 'param: 842\n'

    with MockIO() as mockio:
        CliBuilder(run=sum_param).has(
            parameter('param', multiple=True, type=int),
        ).run()
        assert mockio.output() == 'param: 0\n'


def test_strict_choices():
    with MockIO('--param=4'):
        cli = CliBuilder(reraise_error=True).has(
            parameter('param', choices=['42'], strict_choices=True),
        )
        assert_cli_error(lambda: cli.run())

    with MockIO('--param=42'):
        CliBuilder(reraise_error=True).has(
            parameter('param', choices=['42'], strict_choices=True),
        ).run()

    def complete():
        return ['42']

    with MockIO('--param=42'):
        CliBuilder(reraise_error=True).has(
            parameter('param', choices=complete, strict_choices=True),
        ).run()


def test_multi_default_list():
    def show_param(param):
        print(param)

    with MockIO() as mockio:
        CliBuilder(run=show_param).has(
            parameter('param', multiple=True, type=int, default=[1]),
        ).run()
        assert mockio.stripped() == '[1]'

    with MockIO('--param', '2') as mockio:
        CliBuilder(run=show_param).has(
            parameter('param', multiple=True, type=int, default=[1]),
        ).run()
        assert mockio.stripped() == '[1, 2]'


def test_multi_default_value():
    def show_param(param):
        print(param)

    with MockIO() as mockio:
        CliBuilder(run=show_param).has(
            parameter('param', multiple=True, type=int, default=1),
        ).run()
        assert mockio.stripped() == '[1]'

    with MockIO('--param', '2') as mockio:
        CliBuilder(run=show_param).has(
            parameter('param', multiple=True, type=int, default=1),
        ).run()
        assert mockio.stripped() == '[1, 2]'


def test_parsing_param_syntax_error():
    with MockIO('--param', 'notanumber'):
        cli = CliBuilder(reraise_error=True).has(
            parameter('param', type=int),
        )
        assert_cli_error(lambda: cli.run(), 'parsing parameter "--param":')
    with MockIO('-p', 'notanumber'):
        cli = CliBuilder(reraise_error=True).has(
            parameter('-p', name='param', type=int),
        )
        assert_cli_error(lambda: cli.run(), 'parsing parameter "param":')


def test_strict_multi_parameters():
    with MockIO() as mockio:
        CliBuilder(run=lambda params: print(params)).has(
            parameter('add', name='params', multiple=True, choices=['abc', 'def'], strict_choices=True),
        ).run()
        assert mockio.stripped() == '[]'

    with MockIO('--add', 'def', '--add', 'abc') as mockio:
        CliBuilder(run=lambda params: print(params)).has(
            parameter('add', name='params', multiple=True, choices=['abc', 'def'], strict_choices=True),
        ).run()
        assert mockio.stripped() == "['def', 'abc']"


def test_strict_multi_parameters_fail():
    with MockIO('--add', 'dupa'):
        cli = CliBuilder(run=lambda params: print(params), reraise_error=True).has(
            parameter('add', name='params', multiple=True, choices=['abc', 'def'], strict_choices=True),
        )
        assert_cli_error(lambda: cli.run())



from nuclear import *
from nuclear.types.boolean import boolean
from tests.asserts import MockIO


def test_boolean_type():
    def print_it(param1: bool, param2: bool):
        print(f'{param1} {type(param1).__name__}, {param2} {type(param2).__name__}')

    with MockIO('--param1=true') as mockio:
        CliBuilder(run=print_it).has(
            parameter('param1', type=boolean, default=False),
            parameter('param2', type=boolean, default=True),
        ).run()
        assert mockio.stripped() == 'True bool, True bool'

    with MockIO('--param1=false', '--param2=no') as mockio:
        CliBuilder(run=print_it).has(
            parameter('param1', type=boolean, default=True),
            parameter('param2', type=boolean, default=True),
        ).run()
        assert mockio.stripped() == 'False bool, False bool'



from nuclear.utils.shell import shell, shell_error_code, shell_output
from tests.asserts import assert_error


def test_shell_exec():
    shell('echo test')
    assert_error(lambda: shell('no-faking-command'))
    assert shell_error_code('echo test') == 0
    assert shell_output('echo test') == 'test\n'
    assert shell_output('echo żółć') == u'żółć\n'
    assert shell_output('echo test', as_bytes=True) == b'test\n'
    assert shell_output('echo test', as_bytes=True) == 'test\n'.encode('utf-8')
    assert shell_output('echo test', as_bytes=True).decode('utf-8') == 'test\n'



from .builder.builder import CliBuilder
from .builder.rule_factory import subcommand, flag, dictionary, parameter, primary_option, argument, arguments, \
    default_action
from .version import __version__

name = "nuclear"



import os
import sys


def read_file(file_path):
    with open(file_path, 'rb') as f:
        return f.read().decode('utf-8')


def save_file(file_path, content):
    f = open(file_path, 'wb')
    f.write(bytes(content, 'utf-8'))
    f.close()


def file_exists(path):
    return os.path.isfile(path)


def list_dir(path):
    return sorted(os.listdir(path))


def set_workdir(work_dir):
    os.chdir(work_dir)


def get_workdir():
    return os.getcwd()


def script_real_dir() -> str:
    return os.path.dirname(script_real_path())


def script_real_path() -> str:
    return os.path.realpath(sys.modules['__main__'].__file__)



import re

from .strings import nonempty_lines
from .files import read_file


def regex_match(str_in, regex_match_pattern):
    return bool(re.compile(regex_match_pattern).match(str_in))


def regex_replace(str_in, regex_match_pattern, regex_replace_pattern):
    return re.compile(regex_match_pattern).sub(regex_replace_pattern, str_in)


def regex_filter_list(lines, regex_match_pattern):
    regex_matcher = re.compile(regex_match_pattern)
    return list(filter(lambda line: regex_matcher.match(line), lines))


def regex_replace_list(lines, regex_match_pattern, regex_replace_pattern):
    regex_matcher = re.compile(regex_match_pattern)
    return list(map(lambda line: regex_matcher.sub(regex_replace_pattern, line), lines))


def regex_search_file(file_path, regex_match_pattern, group_number):
    regex_matcher = re.compile(regex_match_pattern)
    with open(file_path) as f:
        for line in f:
            match = regex_matcher.match(line)
            if match:
                return match.group(group_number)


def regex_replace_file(file_path, regex_match_pattern, regex_replace_pattern):
    file_content = read_file(file_path)
    lines = nonempty_lines(file_content)
    lines = regex_replace_list(lines, regex_match_pattern, regex_replace_pattern)
    return '\n'.join(lines)



from typing import Union, Callable, Iterable, List, Type, Any, Optional

Action = Callable[..., None]
ChoiceProvider = Union[Iterable[Any], Callable[[Optional[str]], List[Any]]]
TypeOrParser = Union[Type, Callable[[str], Any]]



from typing import Set


def trim_dashes(arg: str) -> str:
    while arg.startswith('-'):
        arg = arg[1:]
    return arg


def format_var_name(keyword: str) -> str:
    return trim_dashes(keyword).lower().replace('-', '_')


def format_keyword(name: str) -> str:
    name = name.lower().replace('_', '-')
    if name.startswith('-'):
        return name
    if len(name) == 1:
        return '-' + name
    else:
        return '--' + name


def format_var_names(keywords: Set[str]) -> Set[str]:
    return set([format_var_name(k) for k in keywords])


def format_keywords(names: Set[str]) -> Set[str]:
    return set([format_keyword(name) for name in names])



from nuclear import *
from tests.asserts import MockIO
from tests.parser.actions import *


def test_error_on_unrecognized_arg():
    with MockIO('noway', '--noarg') as mockio:
        CliBuilder(run=print_bad, error_unrecognized=True).has(
            subcommand('noway', run=print_bad),
        ).run()
        assert 'bad' not in mockio.output()
        assert 'unrecognized arguments: --noarg\n' in mockio.output()


def test_unrecognized_arg_before_subcommand():
    with MockIO('--noarg', 'noway') as mockio:
        CliBuilder(run=print_bad, error_unrecognized=True).has(
            subcommand('noway', run=print_bad),
        ).run()
        assert 'bad' not in mockio.output()
        assert 'unrecognized arguments: --noarg noway\n' in mockio.output()


def test_ignore_unrecognized_args():
    with MockIO('--noarg', 'noway') as mockio:
        CliBuilder(run=print_ok, error_unrecognized=False).run()
        assert 'ok' in mockio.output()
        assert 'unrecognized arguments: --noarg noway\n' in mockio.output()



from typing import List, Type

from nuclear.builder.rule import CliRule, KeywordRule, TCliRule
from .keyword import format_keywords


def filter_rules(rules: List[CliRule], *types: Type[TCliRule]) -> List[TCliRule]:
    return [r for r in rules if isinstance(r, (*types,))]


def normalize_keywords(rules: List[KeywordRule]):
    for rule in rules:
        rule.keywords = format_keywords(set(rule.keywords))



from nuclear import *
from nuclear.types.filesystem import existing_file, existing_directory
from nuclear.utils.files import script_real_path, script_real_dir
from tests.asserts import MockIO, assert_cli_error


def test_not_existing_file():
    with MockIO('--file', '/dupa'):
        cli = CliBuilder(usage_onerror=False, reraise_error=True).has(
            parameter('file', type=existing_file),
        )
        assert_cli_error(lambda: cli.run())


def test_existing_file():
    script_file = script_real_path()
    with MockIO('--file', script_file):
        CliBuilder(usage_onerror=False, reraise_error=True).has(
            parameter('file', type=existing_file),
        ).run()


def test_not_existing_dir():
    with MockIO('--dir', '/dupa'):
        cli = CliBuilder(usage_onerror=False, reraise_error=True).has(
            parameter('dir', type=existing_directory),
        )
        assert_cli_error(lambda: cli.run())


def test_file_as_dir():
    script_file = script_real_path()
    with MockIO('--dir', script_file):
        cli = CliBuilder(usage_onerror=False, reraise_error=True).has(
            parameter('dir', type=existing_directory),
        )
        assert_cli_error(lambda: cli.run())


def test_existing_dir():
    script_dir = script_real_dir()
    with MockIO('--dir', script_dir):
        CliBuilder(usage_onerror=False, reraise_error=True).has(
            parameter('dir', type=existing_directory),
        ).run()



from typing import List, Any, Set, Optional, TypeVar, Iterable, Union

from dataclasses import dataclass, field

from nuclear.parser.keyword import format_var_names
from .typedef import Action, ChoiceProvider, TypeOrParser


@dataclass
class CliRule(object):
    pass


@dataclass
class HelpRule(CliRule):
    help: str = None


@dataclass
class KeywordRule(CliRule):
    keywords: Set[str] = field(default_factory=lambda: set())


@dataclass
class ParentRule(KeywordRule):
    run: Action = None
    subrules: List['CliRule'] = field(default_factory=lambda: [])

    def has(self, *subrules: CliRule) -> 'ParentRule':
        self.subrules += subrules
        return self


@dataclass
class ValueRule(CliRule):
    name: Optional[str] = None
    type: TypeOrParser = str
    choices: ChoiceProvider = None
    strict_choices: bool = False


@dataclass
class OptionalValueRule(ValueRule):
    required: bool = False
    default: Any = None


@dataclass
class SubcommandRule(ParentRule, HelpRule):
    pass


@dataclass
class PrimaryOptionRule(ParentRule, HelpRule):
    pass


@dataclass
class FlagRule(HelpRule, KeywordRule):
    multiple: bool = False

    def default_value(self) -> Union[bool, int]:
        if self.multiple:
            return 0
        else:
            return False


@dataclass
class DictionaryRule(HelpRule, KeywordRule):
    name: Optional[str] = None
    key_type: TypeOrParser = str
    value_type: TypeOrParser = str

    def var_names(self) -> Iterable[str]:
        if self.name:
            return [self.name]
        else:
            return format_var_names(self.keywords)


@dataclass
class ParameterRule(HelpRule, OptionalValueRule, KeywordRule):
    multiple: bool = False

    def display_name(self) -> str:
        if self.name:
            return self.name
        else:
            return ', '.join(self.keywords)

    def var_names(self) -> Iterable[str]:
        if self.name:
            return [self.name]
        else:
            return format_var_names(self.keywords)

    def default_value(self) -> Any:
        if self.multiple:
            if not self.default:
                return []
            elif not isinstance(self.default, list):
                return [self.default]
            else:
                return self.default
        else:
            return self.default


@dataclass
class PositionalArgumentRule(HelpRule, OptionalValueRule):
    pass


@dataclass
class ManyArgumentsRule(HelpRule, ValueRule):
    count: Optional[int] = None
    min_count: Optional[int] = None
    max_count: Optional[int] = None
    joined_with: Optional[str] = None

    def count_min(self) -> Optional[int]:
        if self.count:
            return self.count
        return self.min_count

    def count_max(self) -> Optional[int]:
        if self.count:
            return self.count
        return self.max_count


@dataclass
class DefaultActionRule(CliRule):
    run: Action = None


TCliRule = TypeVar('TCliRule', bound=CliRule)



import inspect
from collections.abc import Iterable
from typing import List, Any, Optional

from nuclear.builder.rule import ValueRule
from nuclear.builder.typedef import TypeOrParser


def parse_value_rule(rule: ValueRule, arg: str) -> Any:
    return parse_typed_value(rule.type, arg)


def parse_typed_value(_type: TypeOrParser, arg: str) -> Any:
    if _type is None:
        return arg
    # invoke custom parser or cast to custom type
    return _type(arg)


def generate_value_choices(rule: ValueRule, current: Optional[str] = None) -> List[Any]:
    if not rule.choices:
        return []
    elif isinstance(rule.choices, list):
        return rule.choices
    elif isinstance(rule.choices, Iterable):
        return [choice for choice in rule.choices]
    else:
        (args, _, _, _, _, _, annotations) = inspect.getfullargspec(rule.choices)
        if len(args) >= 1:
            results = rule.choices(current=current)
        else:
            results = rule.choices()
        return list(results)



from nuclear import *
from tests.asserts import MockIO


def test_primary_option_cancels_required_args():
    with MockIO('--help') as mockio:
        CliBuilder(with_defaults=True).run()
        assert 'Usage' in mockio.output()
        assert 'Syntax error' not in mockio.output()



from typing import Dict

from nuclear import *
from tests.asserts import MockIO, assert_cli_error


def test_dictionary_setting_value():
    def print_dict(config: Dict):
        print(config)

    with MockIO('--config', 'name', 'value') as mockio:
        CliBuilder(run=print_dict).has(
            dictionary('--config', '-c'),
        ).run()
        assert mockio.stripped() == "{'name': 'value'}"


def test_empty_dictionary():
    def print_dict(config: Dict):
        print(config)

    with MockIO() as mockio:
        CliBuilder(run=print_dict).has(
            dictionary('--config', '-c'),
        ).run()
        assert mockio.stripped() == "{}"


def test_setting_multiple_dict_values():
    def print_dict(config: Dict):
        print(config)

    with MockIO('--config', 'name1', '1', '-c', 'name2', '2') as mockio:
        CliBuilder(run=print_dict).has(
            dictionary('--config', '-c', value_type=int),
        ).run()
        assert mockio.stripped() == "{'name1': 1, 'name2': 2}"


def test_typed_dictionary():
    def print_dict(config: Dict):
        print(config)

    with MockIO('--config', '1', '2') as mockio:
        CliBuilder(run=print_dict).has(
            dictionary('--config', '-c', key_type=int, value_type=int),
        ).run()
        assert mockio.stripped() == "{1: 2}"


def test_explicit_var_name():
    def print_dict(configuration: Dict):
        print(configuration)

    with MockIO('-c', '1', '2') as mockio:
        CliBuilder(run=print_dict).has(
            dictionary('-c', name='configuration'),
        ).run()
        assert mockio.stripped() == "{'1': '2'}"


def test_incomplete_dict_params():
    with MockIO('-c'):
        cli = CliBuilder(reraise_error=True).has(
            dictionary('-c'),
        )
        assert_cli_error(lambda: cli.run(), 'missing key argument for dictionary')
    with MockIO('-c', 'name'):
        cli = CliBuilder(reraise_error=True).has(
            dictionary('-c'),
        )
        assert_cli_error(lambda: cli.run(), 'missing value argument for dictionary')



from typing import List, Union


def nonempty_lines(str_in):
    all_lines = str_in.splitlines()
    return list(filter(lambda l: len(l) > 0, all_lines))


def split_to_tuple(line, attrs_count=None, splitter='\t'):
    parts = line.split(splitter)
    if attrs_count and len(parts) != attrs_count:
        raise RuntimeError(f'invalid split parts count (found: {len(parts)}, expected: {attrs_count}) in line: {line}')
    return tuple(parts)


def split_to_tuples(lines: Union[List[str], str], attrs_count=None, splitter='\t'):
    if not isinstance(lines, list):
        lines = nonempty_lines(lines)
    return list(map(lambda line: split_to_tuple(line, attrs_count, splitter), lines))



from typing import Dict
from typing import Type, Any, List, TypeVar, Optional

from nuclear.args.args_que import ArgsQue
from nuclear.args.container import ArgsContainer
from nuclear.builder.rule import PrimaryOptionRule, ParameterRule, FlagRule, CliRule, KeywordRule, \
    DefaultActionRule, PositionalArgumentRule, ManyArgumentsRule, SubcommandRule, DictionaryRule, ValueRule
from nuclear.builder.typedef import Action
from nuclear.sublog import log
from .context import RunContext
from .error import CliSyntaxError
from .inject import run_action
from .internal_vars import InternalVars
from .matcher import match_param, match_dictionary
from .transform import normalize_keywords, TCliRule, filter_rules
from .validate import validate_rules, check_strict_choices, check_required_arguments
from .value import parse_value_rule, parse_typed_value


class Parser(object):
    def __init__(self,
                 rules: List[CliRule],
                 run: Optional[Action] = None,
                 parent: Optional['Parser'] = None,
                 dry: bool = False,
                 subcommand: Optional[SubcommandRule] = None,
                 error_unrecognized: bool = False,
                 ):
        """
        Command line arguments parser
        :param rules: list of rules defined for the base level of parser
        :param run: default action to invoke when it's triggered
        :param parent: parent parser for sub-parser on the deeper level
        :param dry: whether dry run should be invoked. Dry run does not trigger any action.
        :param subcommand: A subcommand rule from which this parser is derived
        """
        self.__run: Action = run
        self.__rules: List[CliRule] = rules
        self.__subcommand: SubcommandRule = subcommand
        self._init_rules()

        self.__parent: Optional['Parser'] = parent
        self.__action_triggered = parent.__action_triggered if parent else False
        self.__dry = parent.__dry if parent else dry
        self.__error_unrecognized = error_unrecognized

        self.internal_vars = InternalVars()
        self._init_vars()

    def _init_rules(self):
        if not self.__run:
            for rule in self._rules(DefaultActionRule):
                self.__run = rule.run
        normalize_keywords(self._rules(FlagRule, ParameterRule, DictionaryRule))
        validate_rules(self.__rules)

    def _init_vars(self):
        for rule in self._rules(FlagRule):
            for keyword in rule.keywords:
                self.internal_vars[keyword] = rule.default_value()

        for rule in self._rules(ParameterRule):
            for var_name in rule.var_names():
                self.internal_vars[var_name] = rule.default_value()

        for rule in self._rules(PositionalArgumentRule):
            self.internal_vars[rule.name] = rule.default

        for rule in self._rules(ManyArgumentsRule):
            self.internal_vars[rule.name] = []

        for rule in self._rules(DictionaryRule):
            for var_name in rule.var_names():
                self.internal_vars[var_name] = {}

    def parse_args(self, args_list: List[str]) -> Optional[RunContext]:
        """
        Parse arguments list, read all flags, parameters and run triggered actions
        :param args_list:
        :return: run context containing all the details of triggered actions
        and particular state of parser (matched commands)
        """
        args = ArgsQue(args_list[:])
        run_context: Optional[RunContext] = self._parse_args_queue(args)
        self._check_superfluous_args(args)

        if run_context and run_context.action and not self.__dry:
            run_action(run_context.action, run_context.internal_vars)

        return run_context

    def _parse_args_queue(self, args: ArgsQue) -> Optional[RunContext]:
        try:
            self._parse_single_flags(args)
            self._parse_params(args)
            self._parse_dicts(args)
            self._parse_combined_flags(args)
            return self._parse_primary_options(args) or \
                self._parse_subcommand(args) or \
                self._parse_current_level(args)
        except CliSyntaxError as e:
            if self.__dry:
                return self._build_run_context(None)
            else:
                raise e

    def _parse_current_level(self, args: ArgsQue) -> Optional[RunContext]:
        self._parse_positional_arguments(args)
        self._parse_many_arguments(args)
        if not self.__dry:
            self._check_required_arguments()
            self._check_strict_choices()
        return self._run_default_action()

    def _parse_single_flags(self, args: ArgsQue):
        for arg in args:
            rule: FlagRule = self._find_rule(FlagRule, arg)
            if rule:
                args.pop_current()
                self._set_single_flag(rule)

    def _set_single_flag(self, rule: FlagRule):
        for keyword in rule.keywords:
            if rule.multiple:
                oldval = self.internal_vars[keyword]
                self.internal_vars[keyword] = oldval + 1
            else:
                self.internal_vars[keyword] = True

    def _parse_combined_flags(self, args: ArgsQue):
        for arg in args:
            rules = self._extract_combined_flag(arg)
            if rules:
                args.pop_current()
                for rule in rules:
                    self._set_single_flag(rule)

    def _extract_combined_flag(self, arg: str) -> List[FlagRule]:
        if not arg.startswith('-') or arg.startswith('--'):
            return []
        matched_flags = []
        for single_char in arg[1:]:
            rule: FlagRule = self._find_rule(FlagRule, f'-{single_char}')
            if rule:
                matched_flags.append(rule)
            else:
                return []  # every combined character has to be detected as single flag
        return matched_flags

    def _parse_params(self, args: ArgsQue):
        for arg in args:
            for rule in self._rules(ParameterRule):
                value = match_param(rule, args, arg)
                if value:
                    self._parse_param(rule, value)

    def _parse_param(self, rule: ParameterRule, value_str: str):
        try:
            parsed_value = parse_value_rule(rule, value_str)
        except ValueError as e:
            raise CliSyntaxError(f'parsing parameter "{rule.display_name()}"') from e

        self._set_param_value(rule, parsed_value)

    def _set_param_value(self, rule: ParameterRule, parsed_value):
        for var_name in rule.var_names():
            if rule.multiple:
                oldval: List = self.internal_vars[var_name]
                oldval.append(parsed_value)
            else:
                self.internal_vars[var_name] = parsed_value

    def _parse_dicts(self, args: ArgsQue):
        for arg in args:
            for rule in self._rules(DictionaryRule):
                raw_key, raw_value = match_dictionary(rule, args, arg)
                if raw_value:
                    entry_key = parse_typed_value(rule.key_type, raw_key)
                    entry_value = parse_typed_value(rule.value_type, raw_value)
                    self._add_dict_value(rule, entry_key, entry_value)

    def _add_dict_value(self, rule: DictionaryRule, entry_key, entry_value):
        for var_name in rule.var_names():
            oldval: Dict = self.internal_vars[var_name]
            oldval[entry_key] = entry_value
            self.internal_vars[var_name] = oldval

    def _parse_primary_options(self, args: ArgsQue) -> Optional[RunContext]:
        for arg in args:
            rule: PrimaryOptionRule = self._find_rule(PrimaryOptionRule, arg)
            if rule:
                args.pop_current()
                self.__action_triggered = True
                subparser = Parser(rule.subrules, rule.run, parent=self)
                return subparser._parse_args_queue(args)
        return None

    def _parse_subcommand(self, args: ArgsQue) -> Optional[RunContext]:
        if args:
            # recognize first argument as a command
            first = args.reset().peek_current()
            rule: SubcommandRule = self._find_rule(SubcommandRule, first)
            if rule:
                args.pop_current()
                subparser = Parser(rule.subrules, rule.run, parent=self, subcommand=rule)
                return subparser._parse_args_queue(args)
        return None

    def _parse_positional_arguments(self, args: ArgsQue):
        for arg, rule in zip(args.reset(), self._rules(PositionalArgumentRule)):
            try:
                self.internal_vars[rule.name] = parse_value_rule(rule, args.pop_current())
            except ValueError as e:
                raise CliSyntaxError(f'parsing positional argument "{rule.name}"') from e

        if self.__parent and not self.__action_triggered:
            self.__parent._parse_positional_arguments(args)

    def _parse_many_arguments(self, args: ArgsQue):
        args.reset()
        for rule in self._rules(ManyArgumentsRule):
            retrieve_count: Optional[int] = rule.count_max()
            if not retrieve_count:
                retrieve_count = len(args)
            if len(args) < retrieve_count:
                raise CliSyntaxError(f'{retrieve_count} positional arguments are required,'
                                     f' but "{len(args)} given"')
            retrieved = []
            for _ in range(retrieve_count):
                arg = args.pop_current()
                try:
                    parsed_value = parse_value_rule(rule, arg)
                    retrieved.append(parsed_value)
                except ValueError as e:
                    raise CliSyntaxError(f'parsing many arguments "{rule.name}"') from e

            if rule.joined_with:
                var_value = rule.joined_with.join(retrieved)
            else:
                var_value = retrieved
            self.internal_vars[rule.name] = var_value

        if self.__parent:
            self.__parent._parse_many_arguments(args)

    def _run_default_action(self) -> Optional[RunContext]:
        if self.__dry or self.__run:
            return self._build_run_context(self.__run)
        elif self.__parent:
            return self.__parent._run_default_action()
        return None

    def _check_superfluous_args(self, args: ArgsQue):
        if args and not self.__dry:
            if self.__error_unrecognized:
                raise CliSyntaxError(f'unrecognized arguments: {" ".join(args)}')
            else:
                log.warn(f'unrecognized arguments: {" ".join(args)}')

    def _check_required_arguments(self):
        check_required_arguments(self.__rules, self.internal_vars)
        if self.__parent and not self.__action_triggered:
            self.__parent._check_required_arguments()

    def _check_strict_choices(self):
        check_strict_choices(self._rules(ValueRule), self.internal_vars)
        if self.__parent and not self.__action_triggered:
            self.__parent._check_strict_choices()

    def _rules(self, *types: Type[TCliRule]) -> List[TCliRule]:
        return filter_rules(self.__rules, *types)

    TKeywordRule = TypeVar('TKeywordRule', bound=KeywordRule)

    def _find_rule(self, rule_type: Type[TKeywordRule], keyword: str) -> Optional[TKeywordRule]:
        for rule in self.__rules:
            if isinstance(rule, rule_type):
                if keyword in rule.keywords:
                    return rule

    def _internal_vars_merged(self) -> Dict[str, Any]:
        if self.__parent:
            return {**self.__parent._internal_vars_merged(), **self.internal_vars.vars}
        return self.internal_vars.vars

    def _build_run_context(self, action: Optional[Action]) -> RunContext:
        internal_vars = self._internal_vars_merged()
        args_container = ArgsContainer(internal_vars)
        active_subcommands = self._active_subcommands()
        active_rules = self._active_rules()
        return RunContext(args_container, action, active_subcommands, active_rules, internal_vars)

    def _active_subcommands(self) -> List[SubcommandRule]:
        subcommands: List[SubcommandRule] = []
        if self.__subcommand:
            subcommands.append(self.__subcommand)
        if self.__parent:
            return self.__parent._active_subcommands() + subcommands
        return subcommands

    def _active_rules(self) -> List[CliRule]:
        active_rules: List[CliRule] = self.__rules[:]
        if self.__parent:
            active_rules.extend(self.__parent._active_rules())
        return active_rules



from datetime import datetime, date, time
from typing import Callable

from nuclear.parser.error import CliSyntaxError


def datetime_format(*formats: str) -> Callable[[str], datetime]:
    """format: %Y-%m-%d %H:%M:%S"""

    def parser(arg: str):
        return _parse_date_formats(arg, *formats)

    return parser


def today_format(*formats: str) -> Callable[[str], datetime]:
    """format: %Y-%m-%d %H:%M:%S"""

    def parser(arg: str):
        today: date = datetime.now().date()
        parsed_time: time = _parse_date_formats(arg, *formats).time()
        return datetime.combine(today, parsed_time)

    return parser


iso_date: Callable[[str], datetime] = datetime_format('%Y-%m-%d')
iso_time: Callable[[str], datetime] = datetime_format('%H:%M:%S')
iso_datetime: Callable[[str], datetime] = datetime_format('%Y-%m-%d %H:%M:%S')


def _parse_date_formats(s: str, *formats: str) -> datetime:
    for time_format in formats:
        try:
            return _parse_date(s, time_format)
        except ValueError:
            pass
    raise CliSyntaxError('invalid datetime format: ' + s)


def _parse_date(s: str, time_format: str):
    return datetime.strptime(s, time_format)



from nuclear.utils.time import *


def test_time_conversions():
    pattern = '%H:%M:%S, %d.%m.%Y'
    sample_date = '16:26:01, 20.12.2017'
    bad_date = '16:26:01 20.12.17dupa'
    dt = str2time(sample_date, pattern)
    assert dt is not None
    assert time2str(dt, pattern) == sample_date
    assert str2time(bad_date, pattern) is None
    assert time2str(None, pattern) is None



from typing import List

from nuclear import *
from tests.asserts import MockIO, assert_cli_error


def print_all_args(remainder: List[str]):
    print(f'remainder: {", ".join(remainder)}')


def test_remaining_args():
    with MockIO('1', '2', '3') as mockio:
        CliBuilder(run=print_all_args).has(
            arguments('remainder'),
        ).run()
        assert mockio.output() == 'remainder: 1, 2, 3\n'


def test_remaining_args_empty():
    with MockIO() as mockio:
        CliBuilder(run=print_all_args).has(
            arguments('remainder'),
        ).run()
        assert mockio.output() == 'remainder: \n'


def test_pos_arg_with_remaining_args():
    def print_args(arg1, arg2, remainder: List[str]):
        print(f'args: {arg1} {arg2} {", ".join(remainder)}')

    with MockIO('1', '2', '3', '4') as mockio:
        CliBuilder(run=print_args).has(
            argument('arg1'),
            argument('arg2'),
            arguments('remainder'),
        ).run()
        assert mockio.output() == 'args: 1 2 3, 4\n'


def test_remaining_args_joined():
    def proint_remainder(remainder: str):
        print(f'remainder: {remainder}')

    with MockIO('1', '2', '3') as mockio:
        CliBuilder(run=proint_remainder).has(
            arguments('remainder', joined_with=' '),
        ).run()
        assert mockio.output() == 'remainder: 1 2 3\n'


def test_all_argument_in_lower_subcommand():
    def print_components(components):
        print(','.join(components))

    with MockIO('all', '1') as mockio:
        CliBuilder().has(
            subcommand('all', run=print_components),
            arguments('components'),
        ).run()
        assert mockio.output() == '1\n'


def test_exact_many_arguments_count():
    with MockIO('all', '2') as mockio:
        CliBuilder(reraise_error=True, run=lambda components: print(len(components))).has(
            arguments('components', count=2),
        ).run()
        assert mockio.stripped() == '2'
    with MockIO():
        cli = CliBuilder(reraise_error=True).has(
            arguments('components', count=2),
        )
        assert_cli_error(lambda: cli.run())


def test_exact_many_arguments_count_and_superfluous():
    with MockIO('all', '2', '3') as mockio:
        CliBuilder(reraise_error=True, error_unrecognized=False, run=lambda components: print(components)).has(
            arguments('components', count=2),
        ).run()
        assert "['all', '2']" in mockio.stripped()


def test_min_many_arguments_count_syntax_error():
    with MockIO():
        cli = CliBuilder(reraise_error=True).has(
            arguments('components', min_count=1),
        )
        assert_cli_error(lambda: cli.run())


def test_min_many_arguments_count():
    with MockIO('one'):
        CliBuilder(reraise_error=True).has(
            arguments('components', min_count=1),
        ).run()


def test_max_many_arguments_count():
    with MockIO('one', 'two') as mockio:
        CliBuilder(reraise_error=True, error_unrecognized=False, run=lambda words: print(len(words))).has(
            arguments('words', max_count=1),
        ).run()
        assert '1' in mockio.stripped()
    with MockIO('one') as mockio:
        CliBuilder(reraise_error=True, run=lambda words: print(len(words))).has(
            arguments('words', max_count=1),
        ).run()
        assert '1' in mockio.stripped()


def test_parsing_many_arguments_typed():
    with MockIO('1', '41') as mockio:
        CliBuilder(run=lambda numbers: print(sum(numbers))).has(
            arguments('numbers', type=int),
        ).run()
        assert mockio.stripped() == '42'


def test_multiple_many_arguments_rules():
    with MockIO('1', '41', 'abc') as mockio:
        CliBuilder(run=lambda numbers, letters: print(letters + numbers)).has(
            arguments('numbers', count=2),
            arguments('letters'),
        ).run()
        assert mockio.stripped() == "['abc', '1', '41']"


def test_strict_choices():
    with MockIO('4'):
        cli = CliBuilder(reraise_error=True).has(
            arguments('a', choices=['42'], strict_choices=True),
        )
        assert_cli_error(lambda: cli.run())

    with MockIO('42'):
        CliBuilder(reraise_error=True).has(
            arguments('a', choices=['42'], strict_choices=True),
        ).run()

    def complete():
        return ['42']

    with MockIO('42'):
        CliBuilder(reraise_error=True).has(
            arguments('a', choices=complete, strict_choices=True),
        ).run()


def test_parsing_syntax_error():
    with MockIO('notanumber'):
        cli = CliBuilder(reraise_error=True).has(
            arguments('nums', type=int),
        )
        assert_cli_error(lambda: cli.run(), 'parsing many arguments "nums":')



from nuclear.args.args_que import ArgsQue


def test_iterating_all():
    args = ArgsQue(['1', '2', '3'])
    assert args
    assert len(args) == 3
    result = []
    for arg in args:
        result += arg
    assert result == ['1', '2', '3']
    # test iterating 2nd time
    result = []
    for arg in args:
        result += arg
    assert result == ['1', '2', '3']


def test_removing_while_iterating():
    args = ArgsQue(['1', '2', '3'])
    result = []
    for arg in args:
        result += arg
        if arg == '2':
            args.pop_current()
    assert result == ['1', '2', '3']
    result = []
    for arg in args:
        result += arg
    assert result == ['1', '3']


def test_removing_all():
    args = ArgsQue(['1', '2', '3'])
    result = []
    for arg in args:
        result += arg
        args.pop_current()
    assert result == ['1', '2', '3']
    assert not args
    assert len(args) == 0


def test_removing_inside_loop():
    args = ArgsQue(['1', '2', '3', '4'])
    result = []
    for arg in args:
        args.pop_current()
        args.pop_current()
        result += arg
    assert result == ['1', '3']
    assert not args
    assert len(args) == 0


def test_removing_by_3():
    args = ArgsQue(['1', '2', '3', '4', '5', '6'])
    result = []
    for _ in args:
        for _ in range(2):
            part = []
            for _ in range(3):
                part.append(args.pop_current())
            result.append(part)

    assert result == [['1', '2', '3'], ['4', '5', '6']]
    assert not args
    assert len(args) == 0


def test_just_peeking():
    args = ArgsQue(['1', '2'])
    result = []
    for _ in args:
        result.append(args.peek_current())
        result.append(args.pop_current())

    assert args.peek_current() is None
    assert result == ['1', '1', '2', '2']


def test_pop_all():
    args = ArgsQue(['1', '2'])
    assert args.pop_all() == ['1', '2']
    assert args.pop_all() == []



from nuclear import *
from tests.asserts import MockIO, assert_cli_error


def print_pos_arg1(arg1):
    print(f'arg1: {arg1}')


def print_2_args(arg1, arg2):
    print(f'args: {arg1}, {arg2}')


def test_1_positional_arg():
    with MockIO('val') as mockio:
        CliBuilder(run=print_pos_arg1).has(
            argument('arg1'),
        ).run()
        assert mockio.output() == 'arg1: val\n'


def test_2_positional_args():
    with MockIO('val1', 'val2') as mockio:
        CliBuilder(run=print_2_args).has(
            argument('arg1'),
            argument('arg2'),
        ).run()
        assert mockio.output() == 'args: val1, val2\n'
    with MockIO('val1', 'val2') as mockio:
        CliBuilder(run=print_2_args).has(
            argument('arg2'),
            argument('arg1'),
        ).run()
        assert mockio.output() == 'args: val2, val1\n'


def test_optional_argument():
    with MockIO() as mockio:
        CliBuilder(run=print_pos_arg1).has(
            argument('arg1', required=False),
        ).run()
        assert mockio.output() == 'arg1: None\n'


def test_missing_required_argument():
    with MockIO():
        cli = CliBuilder(run=print_pos_arg1, usage_onerror=False, reraise_error=True).has(
            argument('arg1'),
        )
        assert_cli_error(lambda: cli.run())


def test_strict_choices():
    with MockIO('4'):
        cli = CliBuilder(reraise_error=True).has(
            argument('a', choices=['42'], strict_choices=True),
        )
        assert_cli_error(lambda: cli.run())

    with MockIO('42'):
        CliBuilder(reraise_error=True).has(
            argument('a', choices=['42'], strict_choices=True),
        ).run()

    def complete():
        return ['42']

    with MockIO('42'):
        CliBuilder(reraise_error=True).has(
            argument('a', choices=complete, strict_choices=True),
        ).run()


def test_parsing_syntax_error():
    with MockIO('notanumber'):
        cli = CliBuilder(reraise_error=True).has(
            argument('num', type=int),
        )
        assert_cli_error(lambda: cli.run(), 'parsing positional argument "num":')



import sys
from typing import Callable, List, Optional

from nuclear.autocomplete.autocomplete import bash_autocomplete
from nuclear.autocomplete.install import install_bash, install_autocomplete
from nuclear.help.help import print_version, print_help, print_usage
from nuclear.parser.error import CliSyntaxError, CliDefinitionError
from nuclear.parser.parser import Parser
from nuclear.sublog import log
from .rule import DefaultActionRule, CliRule
from .rule_factory import default_action, primary_option, arguments, argument


class CliBuilder(object):
    def __init__(self,
                 name: Optional[str] = None,
                 version: Optional[str] = None,
                 help: Optional[str] = None,
                 run: Optional[Callable[..., None]] = None,
                 with_defaults: bool = True,
                 usage_onerror: bool = True,
                 reraise_error: bool = False,
                 hide_internal: bool = True,
                 help_on_empty: bool = False,
                 error_unrecognized: bool = True,
                 ):
        """
        A builder for Command Line Interface specification
        :param name: name of the application for which the CLI is built
        :param version: application version (displayed in help/version output)
        :param help: short description of application
        :param run: reference for a function which should be the default action for empty arguments list
        :param with_defaults: whether default rules and actions should be added.
        Defaults options are:
        -h, --help: displaying help,
        --version: displaying version,
        --install-bash APP-NAME: installing application in bash with autocompleting,
        --autocomplete [CMDLINE...]: internal action for generating autocompleted proposals to be handled by bash
        :param usage_onerror: wheter usage output should be displayed on syntax error
        :param reraise_error: wheter syntax error should not be caught but reraised instead.
        Enabling this causes stack trace to be flooded to the user.
        :param hide_internal: wheter internal options (--install-bash, --autocomplete)
        should be hidden on help output.
        :param help_on_empty: shows help output when no argument is given
        :param error_unrecognized: raise error when unrecognized argument is found
        """
        self.__name: str = name
        self.__version: str = version
        self.__help: str = help
        self.__subrules: List[CliRule] = []

        if run:
            self.has(default_action(run))

        self.__usage_onerror: bool = usage_onerror
        self.__reraise_error: bool = reraise_error
        self.__hide_internal: bool = hide_internal
        self.__help_on_empty: bool = help_on_empty
        self.__error_unrecognized: bool = error_unrecognized
        if with_defaults:
            self.__add_default_rules()

    def has(self, *subrules: CliRule) -> 'CliBuilder':
        """
        Add more CLI rules for the particular level
        :param subrules: Command Line Interface rules,
        such as: subcommands, flags, parameters, options, arguments, default action
        :return: CliBuilder itself for the further building
        """
        self.__subrules += subrules
        return self

    def run(self):
        """
        Parse all the CLI arguments passed to application.
        Then invoke triggered action which were defined before.
        If actions need some parameters, they will be injected based on the parsed arguments.
        """
        self.run_with_args(sys.argv[1:])

    def run_with_args(self, args: List[str]):
        try:
            parser = self.__create_parser(args)
            parser.parse_args(args)
        except CliDefinitionError as e:
            log.error(f'CLI Definition error: {e}')
            raise e
        except CliSyntaxError as e:
            log.error(f'Syntax error: {e}')
            if self.__usage_onerror:
                self.print_usage()
            if self.__reraise_error:
                raise e

    def __create_parser(self, args: List[str]) -> Parser:
        if not args and self.__help_on_empty:
            def __print_root_help():
                self.print_help([])

            return Parser([default_action(__print_root_help)])

        return Parser(self.__subrules, error_unrecognized=self.__error_unrecognized)

    def print_help(self, subcommands: List[str]):
        print_help(self.__subrules, self.__name, self.__version, self.__help, subcommands, self.__hide_internal)

    def print_usage(self):
        print_usage(self.__subrules)

    def __bash_autocomplete(self, cmdline: str, word_idx: Optional[int]):
        bash_autocomplete(self.__subrules, cmdline, word_idx)

    def __add_default_rules(self):
        def __print_root_help():
            self.print_help([])

        def __print_subcommand_help(subcommands: List[str]):
            self.print_help(subcommands)

        def __print_version():
            print_version(self.__name, self.__version)

        def __install_bash(app_name: str):
            install_bash(app_name)

        def __install_autocomplete(app_name: Optional[str]):
            install_autocomplete(app_name)

        def __bash_autocomplete(cmdline: str, word_idx: Optional[int]):
            self.__bash_autocomplete(cmdline, word_idx)

        if self.__version:
            self.has(
                primary_option('--version', run=__print_version, help='Print version information and exit'),
            )

        self.has(
            primary_option('-h', '--help', run=__print_subcommand_help, help='Display this help and exit').has(
                arguments('subcommands'),
            ),
            primary_option('--install-bash', run=__install_bash,
                           help='Install this program in bash to be executable from anywhere, '
                                'add autocompletion links').has(
                argument('app-name', help='binary name'),
            ),
            primary_option('--install-autocomplete', run=__install_autocomplete,
                           help='Install autocompletion links').has(
                argument('app-name', help='binary name', required=False),
            ),
            primary_option('--autocomplete', run=__bash_autocomplete,
                           help='Return matching autocompletion proposals').has(
                argument('cmdline', help='current command line'),
                argument('word-idx', help='current word index', type=int, required=False),
            ),
        )

        if not self.__has_default_action():
            self.has(default_action(__print_root_help))

    def __has_default_action(self) -> bool:
        return any([isinstance(rule, DefaultActionRule) for rule in self.__subrules])



from nuclear.sublog import log_error, log, context_logger, root_context_logger
from tests.asserts import MockIO


def test_context_logger():
    with MockIO() as mockio:
        with context_logger(request_id=0xdeaddead) as logger:
            logger.debug('got request')
            with context_logger(logger, user='igrek') as logger2:
                logger2.info('logged in', page='sweet home')
                log.warn('im a root')

            logger.debug('logged out')
        log.debug(42)

        # datetime
        mockio.assert_match('^\\[\x1b\\[36m\\d{4}-\\d{2}-\\d{2} \\d{2}:\\d{2}:\\d{2}\x1b\\[0m\\] ')
        # log level
        mockio.assert_match(' \\[\x1b\\[32mDEBUG\x1b\\[0m\\] ')
        # message with context
        mockio.assert_match(' got request \x1b\\[32mrequest_id=\x1b\\[1m3735936685\x1b\\[0m$')

        mockio.assert_match_uncolor('] logged in request_id=3735936685 user=igrek page="sweet home"$')
        mockio.assert_match_uncolor('] im a root$')
        mockio.assert_match_uncolor('] logged out request_id=3735936685$')
        mockio.assert_match_uncolor('] 42$')


def test_root_context_logger():
    with MockIO() as mockio:
        log.debug('outside context', a=4)

        with root_context_logger(request_id=0xdeaddead):
            log.debug('got request')

            with root_context_logger(user='igrek'):
                log.info('logged in', page='home')
                with log_error():
                    log.warn('im a root')
                    raise RuntimeError("I'm a pickle")

            log.debug('logged out')

        log.debug('exited')

        mockio.assert_match_uncolor('] outside context a=4$')
        mockio.assert_match_uncolor('] got request request_id=3735936685$')
        mockio.assert_match_uncolor('] logged in request_id=3735936685 user=igrek page=home$')
        mockio.assert_match_uncolor('] im a root request_id=3735936685 user=igrek$')
        mockio.assert_match_uncolor('] I\'m a pickle request_id=3735936685 user=igrek traceback=.+:40$')
        mockio.assert_match_uncolor('] logged out request_id=3735936685$')
        mockio.assert_match_uncolor('] exited$')



from nuclear.utils.strings import *
from tests.asserts import assert_error


def test_split_lines():
    assert nonempty_lines('a\nb\nc') == ['a', 'b', 'c']
    assert nonempty_lines('\na\n\n') == ['a']
    assert nonempty_lines('\n\n\n') == []
    assert nonempty_lines('') == []
    assert nonempty_lines('a\n\n\r\nb') == ['a', 'b']


def test_split_to_tuple():
    assert split_to_tuple('a', 1) == ('a',)
    assert split_to_tuple('', 1) == ('',)
    assert_error(lambda: split_to_tuple('a', 2))
    assert split_to_tuple('a\tb', 2) == ('a', 'b')
    assert_error(lambda: split_to_tuple('a\tb', 1))
    assert_error(lambda: split_to_tuple('a\tb\t', 2))
    assert split_to_tuple('a\tb\t', 3) == ('a', 'b', '')
    assert split_to_tuple('a\tb\tc', 3) == ('a', 'b', 'c')
    assert split_to_tuple('a b c', 3, ' ') == ('a', 'b', 'c')
    # no attrsCount verification
    assert split_to_tuple('a b c', splitter=' ') == ('a', 'b', 'c')
    assert split_to_tuple('a') == ('a',)


def test_split_to_tuples():
    assert split_to_tuples('a\tb\tc\nd\te\tf', 3) == [('a', 'b', 'c'), ('d', 'e', 'f')]
    assert split_to_tuples('\n\na\tb\tc\n\nd\te\tf\n', 3) == [('a', 'b', 'c'), ('d', 'e', 'f')]
    assert split_to_tuples('a\tb\tc', 3) == [('a', 'b', 'c')]
    # splitted list as input
    assert split_to_tuples(['a\tb\tc', 'd\te\tf'], 3) == [('a', 'b', 'c'), ('d', 'e', 'f')]
    assert split_to_tuples(['a\tb\tc'], 3) == [('a', 'b', 'c')]



from nuclear import *
from nuclear.types.filesystem import *
from tests.parser.actions import *


def test_multilevel_commands_usage():
    CliBuilder('multiapp', version='1.2.3', with_defaults=False).has(
        subcommand('git').has(
            subcommand('help', run=action_print, help='shows help'),
            subcommand('push', run=action_print).has(
                argument('remote'),
                argument('branch', required=False),
            ),
            subcommand('describe', run=action_print).has(
                flag('--tags', help='show tags'),
            ),
            subcommand('checkout', run=action_print).has(
                argument('branch', choices=list_devices, type=str),
                flag('force', '-f'),
            ),
            parameter('--work-tree', type=existing_directory, default='.'),
            parameter('--config-file', type=existing_file),
            parameter('--count', type=int),
            primary_option('-c', run=action_print).has(
                argument('key-value'),
            ),
        ),
        subcommand('nmcli').has(
            subcommand('device').has(
                subcommand('wifi'),
                argument('device_name', choices=list_devices),
            ),
        ),
        subcommand('xrandr').has(
            parameter('output', required=True, choices=['HDMI', 'eDP1']),
            flag('--primary'),
            default_action(action_print),
        ),
        subcommand('docker').has(
            subcommand('exec', run=action_print).has(
                flag('--it'),
                parameter('-u', name='user'),
                arguments(name='cmd', joined_with=' '),
            ),
        ),
        dictionary('-c', '--config'),
        primary_option('--help', '-h', run=action_print),
    ).run()



from typing import Dict, Any

from .keyword import format_var_name


class InternalVars(object):
    def __init__(self):
        self.vars: Dict[str, Any] = {}

    def __getitem__(self, key):
        return self.vars[format_var_name(key)]

    def __setitem__(self, key, item):
        self.vars[format_var_name(key)] = item



from typing import List, Optional

from nuclear.builder.rule import CliRule, PositionalArgumentRule, ManyArgumentsRule, \
    OptionalValueRule, ValueRule, ParameterRule
from nuclear.parser.internal_vars import InternalVars
from .error import CliDefinitionError, CliSyntaxError
from .transform import filter_rules
from .value import generate_value_choices


def validate_rules(all_rules: List[CliRule]):
    for rule in filter_rules(all_rules, OptionalValueRule):
        if rule.required and rule.default:
            raise CliDefinitionError('argument value may be either required or have the default value')

    unlimited_args = 0
    for rule in filter_rules(all_rules, PositionalArgumentRule, ManyArgumentsRule):
        if isinstance(rule, PositionalArgumentRule):
            if unlimited_args:
                raise CliDefinitionError('positional argument can\'t be placed after unlimited many arguments rule')
        elif isinstance(rule, ManyArgumentsRule):
            if not rule.max_count and not rule.count:
                unlimited_args += 1


def check_required_arguments(rules: List[CliRule], internal_vars: InternalVars):
    for rule in filter_rules(rules, ParameterRule):
        if rule.required:
            for name in rule.var_names():
                if not internal_vars[name]:
                    raise CliSyntaxError(f'required parameter "{", ".join(rule.keywords)}" is not given')

    for rule in filter_rules(rules, PositionalArgumentRule):
        if rule.required:
            if not internal_vars[rule.name]:
                raise CliSyntaxError(f'required positional argument "{rule.name}" is not given')

    for rule in filter_rules(rules, ManyArgumentsRule):
        expected_count: Optional[int] = rule.count_min()
        if expected_count:
            given_count = len(internal_vars[rule.name])
            if given_count != expected_count:
                raise CliSyntaxError(f'"{expected_count}" arguments are required, but "{given_count}" were given')


def check_strict_choices(rules: List[ValueRule], internal_vars: InternalVars):
    for rule in rules:
        if rule.strict_choices and rule.choices:
            available_choices = generate_value_choices(rule)

            if isinstance(rule, ParameterRule):
                for name in rule.var_names():
                    if rule.multiple:
                        for var_value in internal_vars[name]:
                            if var_value not in available_choices:
                                raise CliSyntaxError(
                                    f'parameter value {var_value} does not belong to available choices: '
                                    f'{available_choices}')
                    else:
                        var_value = internal_vars[name]
                        if var_value not in available_choices:
                            raise CliSyntaxError(f'parameter value {var_value} does not belong to available choices: '
                                                 f'{available_choices}')

            elif isinstance(rule, PositionalArgumentRule):
                var_value = internal_vars[rule.name]
                if var_value not in available_choices:
                    raise CliSyntaxError(
                        f'positional argument value {var_value} does not belong to available choices: '
                        f'{available_choices}')

            elif isinstance(rule, ManyArgumentsRule):
                var_values: List = internal_vars[rule.name]
                for var_value in var_values:
                    if var_value not in available_choices:
                        raise CliSyntaxError(
                            f'one of arguments value {var_value} does not belong to available choices: '
                            f'{available_choices}')



from nuclear import *
from tests.asserts import MockIO


def test_empty_builder():
    with MockIO() as mockio:
        CliBuilder()
        assert mockio.output() == ''
    with MockIO():
        CliBuilder(with_defaults=False).run()



from nuclear import *
from tests.asserts import MockIO
from tests.parser.actions import *


def test_run_selected_command():
    with MockIO('push') as mockio:
        CliBuilder().has(
            subcommand('commit', run=print_bad),
            subcommand('push', run=print_ok),
            subcommand('checkout', run=print_bad),
        ).run()
        assert mockio.output() == 'ok\n'


def test_2nd_level_subcommand():
    with MockIO('git', 'remote') as mockio:
        CliBuilder().has(
            subcommand('git', run=print_bad).has(
                subcommand('remote', run=print_ok),
            ),
        ).run()
        assert mockio.output() == 'ok\n'


def test_run_selected_subcommand():
    with MockIO('git', 'remote', 'set-url') as mockio:
        CliBuilder().has(
            subcommand('git', run=print_bad).has(
                subcommand('push', run=print_bad),
                subcommand('remote', run=print_bad).has(
                    subcommand('add', run=print_bad),
                    subcommand('set-url', run=print_ok),
                ),
            ),
            subcommand('nmcli', run=print_bad),
        ).run()
        assert mockio.output() == 'ok\n'



import datetime

from nuclear import *
from nuclear.types.time import datetime_format, today_format
from tests.asserts import MockIO, assert_cli_error


def test_datetime_format_generator():
    def print_datetime(d):
        print(str(d))

    with MockIO('-d=2019-06-07') as mockio:
        CliBuilder(run=print_datetime).has(
            parameter('d', type=datetime_format('%Y-%m-%d %H:%M:%S', '%Y-%m-%d')),
        ).run()
        assert mockio.stripped() == str(datetime.datetime(2019, 6, 7))


def test_today_format_generator():
    def print_datetime(d):
        print(str(d))

    with MockIO('-d=12:30') as mockio:
        CliBuilder(run=print_datetime).has(
            parameter('d', type=today_format('%H:%M:%S', '%H:%M')),
        ).run()
        now = datetime.datetime.now()
        assert mockio.stripped() == str(datetime.datetime(now.year, now.month, now.day, 12, 30, 0))


def test_invalid_datetime():
    def print_datetime(d):
        print(str(d))

    with MockIO('-d=2019'):
        cli = CliBuilder(run=print_datetime, usage_onerror=False, reraise_error=True).has(
            parameter('d', type=datetime_format('%Y-%m-%d %H:%M:%S', '%Y-%m-%d')),
        )
        assert_cli_error(lambda: cli.run())



from nuclear import *
from tests.asserts import MockIO
from tests.parser.actions import *


def test_run_default_single_action():
    with MockIO() as mockio:
        CliBuilder(run=print_ok).run()
        assert mockio.output() == 'ok\n'


def test_default_builder_action_with_redundant_args():
    with MockIO('push') as mockio:
        CliBuilder(run=print_ok, error_unrecognized=False).has(
            subcommand('commit', run=print_bad),
            subcommand('checkout', run=print_bad),
        ).run()
        assert 'ok\n' in mockio.output()
    with MockIO('push') as mockio:
        CliBuilder(error_unrecognized=False).has(
            subcommand('commit', run=print_bad),
            subcommand('checkout', run=print_bad),
            default_action(run=print_ok),
        ).run()
        assert 'ok\n' in mockio.output()
    with MockIO('push') as mockio:
        CliBuilder(run=print_bad, error_unrecognized=True).has(
            subcommand('commit', run=print_bad),
            subcommand('checkout', run=print_bad),
        ).run()
        assert 'bad' not in mockio.output()
        assert 'Syntax error: unrecognized arguments' in mockio.output()


def test_subcommand_default_action():
    with MockIO('git', 'push') as mockio:
        CliBuilder(run=print_bad).has(
            subcommand('nmcli').has(
                default_action(run=print_bad)
            ),
            subcommand('git', run=print_bad).has(
                subcommand('push').has(
                    default_action(run=print_ok)
                ),
            ),
        ).run()
        assert mockio.output() == 'ok\n'


def test_parent_default_action():
    with MockIO('nmcli', 'dev') as mockio:
        CliBuilder(run=print_ok).has(
            subcommand('nmcli').has(
                subcommand('dev')
            ),
        ).run()
        assert mockio.output() == 'ok\n'


def test_on_empty_action_overriden():
    with MockIO() as mockio:
        CliBuilder(run=print_bad, help_on_empty=True).has(
            flag('any')
        ).run()
        assert 'Usage' in mockio.output()
    with MockIO('--any') as mockio:
        CliBuilder(run=print_ok, help_on_empty=True).has(
            flag('any')
        ).run()
        assert mockio.output() == 'ok\n'



from typing import Dict, Any

from nuclear.parser.keyword import format_var_name


class ArgsContainer(object):
    def __init__(self, _vars: Dict[str, Any]):
        self.__vars = _vars

    def __getattr__(self, name):
        if name in self.__vars:
            return self.__vars[name]
        else:
            raise AttributeError

    def __getitem__(self, key):
        if key in self.__vars:
            return self.__vars[key]
        else:
            return self.__vars[format_var_name(key)]



from nuclear.sublog import wrap_context, log_error, ContextError, log
from tests.asserts import MockIO


def test_sublog_wrapping():
    with MockIO() as mockio:
        with log_error():
            with wrap_context('initializing', request_id=42):
                with wrap_context('liftoff', speed='zero'):
                    raise RuntimeError('dupa')

        with log_error():
            raise ContextError('dupa2', a=5, z='fifteen')

        with log_error():
            raise RuntimeError('dupa3')

        log.info('success', param='with_param')
        log.warn('attention')
        log.debug('trace')

        mockio.assert_match_uncolor('] initializing: liftoff: dupa '
                                    'request_id=42 speed=zero '
                                    'traceback=(.*)/test_error_wrap.py:10$')
        mockio.assert_match_uncolor('] dupa2 a=5 z=fifteen '
                                    'traceback=(.*)/test_error_wrap.py:13$')
        mockio.assert_match_uncolor('\\[ERROR] dupa3 '
                                    'traceback=(.*)/test_error_wrap.py:16$')
        mockio.assert_match_uncolor('\\[INFO ] success param=with_param$')
        mockio.assert_match_uncolor('\\[WARN ] attention$')
        mockio.assert_match_uncolor('\\[DEBUG] trace$')


def test_sublog_traceback():
    with MockIO() as mockio:
        with log_error():
            with wrap_context('initializing', request_id=42):
                with wrap_context('liftoff', speed='zero'):
                    disaster()

        mockio.assert_match_uncolor('ERROR] initializing: liftoff: disaster request_id=42 speed=zero '
                                    'traceback=(.+)/test_error_wrap.py:39,'
                                    '(.+)/test_error_wrap.py:48,'
                                    '(.+)/test_error_wrap.py:52$')


def disaster():
    reason()


def reason():
    raise RuntimeError('disaster')


def test_keyboard_interrupt():
    with MockIO() as mockio:
        try:
            with log_error():
                raise KeyboardInterrupt()
            assert False, 'should exit'
        except SystemExit as e:
            assert str(e) == '1', 'should exit with error code 1'

        assert 'KeyboardInterrupt' in mockio.output()



import datetime
from contextlib import contextmanager
from typing import Dict, Any, Optional

from colorama import Fore, Style


class ContextLogger(object):
    def __init__(self, ctx: Dict[str, Any]):
        self.ctx: Dict[str, Any] = ctx

    def error(self, message: str, **ctx):
        self._print_log(message, f'[{Fore.RED + Style.BRIGHT}ERROR{Style.RESET_ALL}]', ctx)

    def warn(self, message: str, **ctx):
        self._print_log(message, f'[{Fore.YELLOW + Style.BRIGHT}WARN{Style.RESET_ALL} ]', ctx)

    def info(self, message: str, **ctx):
        self._print_log(message, f'[{Fore.BLUE}INFO{Style.RESET_ALL} ]', ctx)

    def debug(self, message: str, **ctx):
        self._print_log(message, f'[{Fore.GREEN}DEBUG{Style.RESET_ALL}]', ctx)

    def _print_log(self, message: str, level: str, ctx: Dict[str, Any]):
        merged_context = {**self.ctx, **ctx}
        display_context = _display_context(merged_context)
        timestamp_part = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        if display_context:
            print(f'[{Fore.CYAN}{timestamp_part}{Style.RESET_ALL}] {level} {message} {display_context}')
        else:
            print(f'[{Fore.CYAN}{timestamp_part}{Style.RESET_ALL}] {level} {message}')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


"""Root logger"""
log = ContextLogger(dict())


def context_logger(parent: Optional[ContextLogger] = None, **ctx) -> ContextLogger:
    if parent is None:
        return ContextLogger(ctx)
    else:
        merged_context = {**parent.ctx, **ctx}
        return ContextLogger(merged_context)


@contextmanager
def root_context_logger(**ctx) -> ContextLogger:
    ctx_backup: Dict[str, Any] = log.ctx
    try:
        log.ctx = {**ctx_backup, **ctx}
        yield log
    finally:
        log.ctx = ctx_backup


def _display_context(ctx: Dict[str, Any]) -> str:
    if len(ctx) == 0:
        return ''
    keys = ctx.keys()
    parts = [_display_context_var(key, ctx[key]) for key in keys]
    return " ".join(parts)


def _display_context_var(var: str, val: str) -> str:
    val = str(val)
    if ' ' in val:
        return f'{Fore.GREEN}{var}="{Style.BRIGHT}{val}{Style.RESET_ALL}{Fore.GREEN}"{Style.RESET_ALL}'
    else:
        return f'{Fore.GREEN}{var}={Style.BRIGHT}{val}{Style.RESET_ALL}'



from nuclear.utils.files import *


def test_read_file():
    assert read_file('tests/utils/res/readme') == 'Readme\n123'


def test_save_file():
    save_file('tests/utils/res/saveme', 'dupa\n123')
    assert read_file('tests/utils/res/saveme') == 'dupa\n123'
    save_file('tests/utils/res/saveme', '')
    assert read_file('tests/utils/res/saveme') == ''


def test_list_dir():
    assert list_dir('tests/utils/res/listme') == ['afile', 'dir', 'zlast', 'zlastdir']


def test_workdir():
    workdir = get_workdir()
    set_workdir('/')
    assert get_workdir() == '/'
    set_workdir('/home/')
    assert get_workdir() == '/home'
    set_workdir(workdir)


def test_script_real_path():
    assert os.path.isfile(script_real_path())
    assert '/pytest.py' in script_real_path() or '/pytest/__main__.py' in script_real_path()


def test_file_exists():
    assert file_exists('tests/utils/res/readme')
    assert not file_exists('tests/utils/res/dupadupa')



from dataclasses import dataclass
from typing import Optional, List, Any, Dict

from nuclear.args.container import ArgsContainer
from nuclear.builder.rule import SubcommandRule, CliRule
from nuclear.builder.typedef import Action


@dataclass
class RunContext(object):
    args_container: ArgsContainer
    action: Optional[Action]
    active_subcommands: List[SubcommandRule]
    active_rules: List[CliRule]
    internal_vars: Dict[str, Any]



import os
from pathlib import Path

from nuclear import *
from nuclear.utils.files import script_real_path
from nuclear.utils.shell import shell
from tests.asserts import MockIO


def test_bash_install_twice():
    app_name = 'nuclear_test_dupa123'

    with MockIO('--install-bash', 'nuclear_test_dupa123') as mockio:
        CliBuilder().run()
        assert 'Link installed' in mockio.output()
        assert 'Autocompleter has been installed' in mockio.output()
        assert os.path.islink(f'/usr/bin/{app_name}')
        assert os.path.realpath(f'/usr/bin/{app_name}') == script_real_path()
        assert os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')

    with MockIO('--install-bash', 'nuclear_test_dupa123') as mockio:
        CliBuilder().run()
        assert 'Link installed' in mockio.output()
        assert 'Autocompleter has been installed' in mockio.output()
        assert os.path.islink(f'/usr/bin/{app_name}')
        assert os.path.realpath(f'/usr/bin/{app_name}') == script_real_path()
        assert os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')

    shell(f'sudo rm -f /usr/bin/{app_name}')
    shell(f'sudo rm -f /etc/bash_completion.d/nuclear_{app_name}.sh')
    assert not os.path.exists(f'/usr/bin/{app_name}')
    assert not os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')


def test_autocomplete_install_explicit_name():
    app_name = 'nuclear_test_dupa123'

    with MockIO('--install-autocomplete', 'nuclear_test_dupa123') as mockio:
        CliBuilder().run()
        assert 'Autocompleter has been installed' in mockio.output()
        assert os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')
        completion_script = Path(f'/etc/bash_completion.d/nuclear_{app_name}.sh').read_text()
        assert '''COMPREPLY=($(nuclear_test_dupa123 --autocomplete "${COMP_LINE}" ${COMP_CWORD}))''' \
               in completion_script
        assert '''complete -o filenames -F _autocomplete_1446250409 nuclear_test_dupa123''' in completion_script

    shell(f'sudo rm -f /etc/bash_completion.d/nuclear_{app_name}.sh')
    assert not os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')


def test_autocomplete_install_implicit_name():
    app_name = 'glue'

    with MockIO('--install-autocomplete') as mockio:
        CliBuilder().run()
        assert 'Autocompleter has been installed' in mockio.output()
        assert os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')
        completion_script = Path(f'/etc/bash_completion.d/nuclear_{app_name}.sh').read_text()
        assert '''COMPREPLY=($(glue --autocomplete "${COMP_LINE}" ${COMP_CWORD}))''' in completion_script
        assert '''complete -o filenames -F _autocomplete_70451630 glue''' in completion_script

    shell(f'sudo rm -f /etc/bash_completion.d/nuclear_{app_name}.sh')
    assert not os.path.exists(f'/etc/bash_completion.d/nuclear_{app_name}.sh')



from nuclear import *
from nuclear.parser.error import CliDefinitionError
from tests.asserts import assert_error


def test_multilevel_commands_usage():
    cli = CliBuilder().has(
        parameter('d', required=True, default='42'),
    )
    assert_error(lambda: cli.run(), error_type=CliDefinitionError)



import os
from typing import Optional, Tuple


def file_completer(current: Optional[str]):
    listdir, prefixdir = _current_listing_dir(current)
    names = []
    for file in os.listdir(listdir):
        filepath = f'{prefixdir}{file}'
        if os.path.isdir(filepath):
            names.append(f'{filepath}')
        else:
            names.append(filepath)
    return sorted(names)


def _current_listing_dir(current: Optional[str]) -> Tuple[str, str]:
    if not current:
        return '.', ''

    current_path, current_node = os.path.split(current)
    if not current_path or current_path == '.':
        return '.', ''

    if not os.path.isdir(current_path):
        return '.', ''

    return current_path, f'{current_path}/'



from nuclear import *
from tests.asserts import MockIO


def test_custom_parser():
    def my_parser(arg):
        return eval(arg)

    def print_my(my):
        print(my)

    with MockIO('--my=17+5') as mockio:
        CliBuilder(run=print_my).has(
            parameter('my', type=my_parser),
        ).run()
        assert mockio.stripped() == '22'


def test_type_none():
    with MockIO('--param', 'notanumber') as mockio:
        CliBuilder(reraise_error=True, run=lambda param: print(param)).has(
            parameter('param', type=None),
        ).run()
        assert 'notanumber' in mockio.stripped()



import inspect
from typing import Any, List
from typing import Dict, Mapping

from nuclear.args.container import ArgsContainer
from nuclear.builder.typedef import Action
from nuclear.sublog import log


def run_action(action: Action, internal_vars: Dict[str, Any]):
    if action:
        (args, _, _, _, _, _, annotations) = inspect.getfullargspec(action)
        if not args:
            action()
        else:
            kwargs = inject_args(args, action, annotations, internal_vars)
            action(**kwargs)


def inject_args(args: List[str], action: Action, annotations: Mapping[str, Any], internal_vars: Dict[str, Any]
                ) -> Dict[str, Any]:
    return {arg: inject_arg(arg, action, annotations, internal_vars) for arg in args}


def inject_arg(arg: str, action: Action, annotations: Mapping[str, Any], internal_vars: Dict[str, Any]) -> Any:
    if arg in internal_vars:
        return internal_vars[arg]
    elif is_args_container_name(arg, annotations):
        return ArgsContainer(internal_vars)
    else:
        log.warn(f"can't inject argument '{arg}' to function '{action.__name__}': name not found")
        return None


def is_args_container_name(arg: str, annotations: Mapping[str, Any]):
    if arg == 'args':
        if arg in annotations:  # has type annotation
            return annotations[arg] == ArgsContainer
        return True
    return False



import re
import sys
from io import StringIO
from typing import Type

import mock

from nuclear.parser.error import CliError


def assert_error(action, error_type: Type[Exception] = RuntimeError, expected_msg: str = None):
    try:
        action()
        assert False, 'should raise error'
    except error_type as e:
        assert isinstance(e, error_type)
        if expected_msg:
            assert expected_msg in str(e), 'unexpected error message'


def assert_cli_error(action, expected_error: str = None):
    assert_error(action, CliError, expected_error)


def assert_system_exit(action):
    try:
        action()
        assert False
    except SystemExit as e:
        # exit with error code 0
        assert str(e) == '0'


class MockIO:
    def __init__(self, *in_args: str):
        # mock cli input
        self._mock_args = mock.patch.object(sys, 'argv', ['glue'] + list(in_args))
        # mock output
        self.new_out, self.new_err = StringIO(), StringIO()
        self.old_out, self.old_err = sys.stdout, sys.stderr

    def __enter__(self):
        self._mock_args.__enter__()
        sys.stdout, sys.stderr = self.new_out, self.new_err
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._mock_args.__exit__(exc_type, exc_value, traceback)
        sys.stdout, sys.stderr = self.old_out, self.old_err
        sys.stdout.write(self.output())

    def output(self) -> str:
        return self.new_out.getvalue()

    def stripped(self) -> str:
        return self.output().strip()

    def uncolor(self) -> str:
        matcher = re.compile(r'\x1b\[[0-9]+(;[0-9]+)?m')
        return matcher.sub('', self.output())

    def assert_match(self, regex: str):
        matcher = re.compile(regex)
        for line in self.output().splitlines():
            if matcher.search(line):
                return
        assert False, f'Regex: "{regex}" does not match the output: {self.output()}'

    def assert_match_uncolor(self, regex: str):
        matcher = re.compile(regex)
        for line in self.uncolor().splitlines():
            if matcher.search(line):
                return
        assert False, f'Regex: "{regex}" does not match the output: {self.uncolor()}'

    def assert_not_match(self, regex: str):
        matcher = re.compile(regex)
        for line in self.output().splitlines():
            if matcher.search(line):
                assert False, f'Regex: "{regex}" should not match the output: {self.output()}'



import os

from nuclear.parser.error import CliSyntaxError


def existing_file(arg: str) -> str:
    if not os.path.isfile(arg):
        raise CliSyntaxError('file does not exist: ' + arg)
    return arg


def existing_directory(arg: str) -> str:
    if not os.path.isdir(arg):
        raise CliSyntaxError('directory does not exist: ' + arg)
    return arg



from nuclear import *
from tests.asserts import MockIO, assert_cli_error


def test_error_pos_arg_after_unlimited_many_args():
    with MockIO():
        cli = CliBuilder(reraise_error=True).has(
            arguments('unlimited'),
            argument('next'),
        )
        assert_cli_error(lambda: cli.run())



import subprocess
from typing import Union


def shell(cmd: str):
    err_code = shell_error_code(cmd)
    if err_code != 0:
        raise RuntimeError(f'command failed: {cmd}')


def shell_error_code(cmd: str) -> int:
    return subprocess.call(cmd, shell=True)


def shell_output(cmd: str, as_bytes: bool = False) -> Union[str, bytes]:
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    output, _ = process.communicate()
    if as_bytes:
        return output
    else:
        return output.decode('utf-8')



import os
import sys
import zlib
from typing import Optional

from nuclear.sublog import log
from nuclear.utils.files import script_real_path
from nuclear.utils.shell import shell


def install_bash(app_name: str):
    """
    Install script link in /usr/bin/{app_name}
    and create bash autocompletion script
    """
    app_path: str = script_real_path()
    assert os.path.isfile(app_path)

    if os.geteuid() != 0:
        log.warn("gaining root privileges in order to install script")

    # ensure script is executable
    if not os.access(app_path, os.X_OK):
        log.info(f'making script executable')
        shell(f'chmod +x {app_path}')

    # creating /usr/bin/ link
    usr_bin_executable: str = f'/usr/bin/{app_name}'
    if os.path.exists(usr_bin_executable) or os.path.islink(usr_bin_executable):
        log.warn(f'file {usr_bin_executable} already exists - skipping.')
    else:
        log.info(f'creating link: {usr_bin_executable} -> {app_path}')
        shell(f'sudo ln -s {app_path} {usr_bin_executable}')

    log.info(f'Link installed in {usr_bin_executable}. Please restart your shell.')
    install_autocomplete(app_name)


def install_autocomplete(app_name: Optional[str]):
    """
    Create bash autocompletion script
    """
    if os.geteuid() != 0:
        log.warn("gaining root privileges in order to install autocompletion script")

    if not app_name:
        app_name = shell_command_name()

    completion_script_path: str = f'/etc/bash_completion.d/nuclear_{app_name}.sh'
    app_hash: int = zlib.adler32(app_name.encode('utf-8'))
    # function should be unique across bash env
    function_name: str = f'_autocomplete_{app_hash}'
    shell(f"""cat << 'EOF' | sudo tee {completion_script_path}
#!/bin/bash
{function_name}() {{
    IFS=$'\n'
    COMPREPLY=($({app_name} --autocomplete "${{COMP_LINE}}" ${{COMP_CWORD}}))
}}
complete -o filenames -F {function_name} {app_name}
EOF
""")
    log.info(f'Autocompleter has been installed in {completion_script_path} for command "{app_name}". '
             f'Please restart your shell.')


def shell_command_name():
    _, command = os.path.split(sys.argv[0])
    return command



import os
import sys
import traceback
from contextlib import contextmanager
from typing import Dict, Any

from .context_logger import log


class ContextError(RuntimeError):
    def __init__(self, message: str, **ctx):
        super().__init__(message)
        self.ctx = ctx


@contextmanager
def wrap_context(context_name: str, **ctx):
    try:
        yield
    except ContextError as e:
        merged_context = {**ctx, **e.ctx}
        raise ContextError(f'{context_name}: {e}', **merged_context) from e
    except Exception as e:
        raise ContextError(f'{context_name}: {e}', **ctx) from e


@contextmanager
def log_error(print_traceback: bool = True):
    """Deprecated, use logerr instead"""
    try:
        yield
    except KeyboardInterrupt:
        print()
        log.debug('KeyboardInterrupt')
        exit(1)
    except ContextError as e:
        _print_error_context(str(e), e.ctx, print_traceback)
    except Exception as e:
        _print_error_context(str(e), {}, print_traceback)


@contextmanager
def logerr(print_traceback: bool = True):
    """Catches all exceptions and displays traceback in pretty, concise format"""
    try:
        yield
    except KeyboardInterrupt:
        print()
        log.debug('KeyboardInterrupt')
        exit(1)
    except ContextError as e:
        _print_error_context(str(e), e.ctx, print_traceback)
    except Exception as e:
        _print_error_context(str(e), {}, print_traceback)


def _print_error_context(message: str, ctx: Dict[str, Any], print_traceback: bool):
    if print_traceback:
        ex_type, ex, tb = sys.exc_info()

        # got from traceback.format_exception(ex_type, ex, tb)
        t1 = traceback.TracebackException(type(ex_type), ex, tb, limit=None)
        while t1.__cause__ is not None:
            t1 = t1.__cause__

        frames = traceback.extract_tb(t1.exc_traceback)

        # hide traceback from this file
        lines = [f'{os.path.normpath(frame.filename)}:{frame.lineno}' for frame in frames
                 if not os.path.samefile(frame.filename, __file__)]

        tb = ','.join(lines)
        ctx['traceback'] = tb
    log.error(message, **ctx)



import datetime


def str2time(time_raw, pattern):
    """pattern: %Y-%m-%d, %H:%M:%S"""
    try:
        return datetime.datetime.strptime(time_raw, pattern)
    except ValueError as _:
        return None


def time2str(date_time, pattern):
    """pattern: %Y-%m-%d, %H:%M:%S"""
    if not date_time:
        return None
    return date_time.strftime(pattern)



def action_print():
    print('It works!')


def print_ok():
    print('ok')


def print_bad():
    print('bad')


def print_flag_force(force: bool):
    print(f'force: {force}')


def print_param(param):
    print(f'param: {param}')


def list_devices():
    return ['sda', 'sdb', 'sdc']
